"""Unit Test for otx.algorithms.action.tasks.train."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import pytest
import torch
from mmcv.utils import Config

from otx.algorithms.action.configs.base.configuration import ActionConfig
from otx.algorithms.action.tasks.train import ActionTrainTask
from otx.api.configuration import ConfigurableParameters
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.metrics import (
    BarMetricsGroup,
    LineMetricsGroup,
    Performance,
    ScoreMetric,
)
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.model_template import InstantiationType, TaskFamily, TaskType
from otx.api.entities.task_environment import TaskEnvironment
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    MockModelTemplate,
    generate_action_cls_otx_dataset,
    generate_labels,
)


class TestActionTrainTask:
    """Test class for ActionTrainTask class."""

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.video_len = 3
        self.frame_len = 3

        cls_labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.cls_label_schema = LabelSchemaEntity()
        cls_label_group = LabelGroup(
            name="labels",
            labels=cls_labels,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        self.cls_label_schema.add_group(cls_label_group)
        cls_template = MockModelTemplate(
            model_template_id="template_id",
            model_template_path="template_path",
            name="template",
            task_family=TaskFamily.VISION,
            task_type=TaskType.ACTION_CLASSIFICATION,
            instantiation=InstantiationType.CLASS,
        )
        self.cls_task_environment = TaskEnvironment(
            model=None,
            hyper_parameters=ConfigurableParameters(header="h-params"),
            label_schema=self.cls_label_schema,
            model_template=cls_template,
        )

        self.cls_dataset = generate_action_cls_otx_dataset(self.video_len, self.frame_len, cls_labels)
        self.cls_task = ActionTrainTask(task_environment=self.cls_task_environment)

    @e2e_pytest_unit
    def test_save_model(self, mocker) -> None:
        """Test save_model function."""

        mocker.patch("otx.algorithms.action.tasks.train.torch.load", return_value={"state_dict": None})

        config = ModelConfiguration(ActionConfig(), self.cls_label_schema)
        output_model = ModelEntity(self.cls_dataset, config)
        self.cls_task.save_model(output_model)
        assert output_model.get_data("weights.pth") is not None
        assert output_model.get_data("label_schema.json") is not None
        assert output_model.precision == self.cls_task._precision

    @e2e_pytest_unit
    def test_cancel_training(self) -> None:
        """Test cance_trainng function."""

        class _MockCanceInterface:
            def cancel(self):
                raise RuntimeError("Checking for calling this function")

        self.cls_task.cancel_training()
        assert self.cls_task.reserved_cancel is True

        self.cls_task.cancel_interface = _MockCanceInterface()
        with pytest.raises(RuntimeError):
            self.cls_task.cancel_training()

    @e2e_pytest_unit
    def test_train(self, mocker) -> None:
        """Test train function."""

        config = ModelConfiguration(ActionConfig(), self.cls_label_schema)
        output_model = ModelEntity(self.cls_dataset, config)
        self.cls_task._should_stop = True
        self.cls_task.train(self.cls_dataset, output_model)
        assert self.cls_task._should_stop is False
        assert self.cls_task._is_training is False

        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._init_task", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._train_model", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._get_output_model", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._get_final_eval_results", return_value=0.5)
        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask.save_model", return_value=True)

        self.cls_task._recipe_cfg = None
        self.cls_task._should_stop = False
        with pytest.raises(Exception):
            self.cls_task.train(self.cls_dataset, output_model)
        self.cls_task._recipe_cfg = Config()
        self.cls_task.train(self.cls_dataset, output_model)
        assert output_model.performance == 0.5
        assert self.cls_task._is_training is False

    @e2e_pytest_unit
    def test_train_model(self, mocker) -> None:
        """Test _train_model function."""

        with pytest.raises(Exception):
            self.cls_task._train_model(self.cls_dataset)

        self.cls_task._recipe_cfg = Config({"work_dir": self.cls_task._output_path})
        self.cls_task._model = torch.nn.Module()

        def _mock_train_model(*args, **kwargs):
            with open(os.path.join(self.cls_task._recipe_cfg.work_dir, "best.pth"), "wb") as f:
                torch.save(torch.randn(1), f)

        mocker.patch(
            "otx.algorithms.action.tasks.train.prepare_for_training", return_value=Config({"data": {"train": None}})
        )
        mocker.patch("otx.algorithms.action.tasks.train.build_dataset", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.train.train_model", side_effect=_mock_train_model)

        out = self.cls_task._train_model(self.cls_dataset)
        assert out["final_ckpt"] is not None
        assert out["final_ckpt"].split("/")[-1] == "best.pth"

    @e2e_pytest_unit
    def test_get_output_model(self, mocker) -> None:
        """Test _get_output_model function."""

        self.cls_task._model = torch.nn.Module()
        sample_results = {"final_ckpt": None}
        self.cls_task._get_output_model(sample_results)

        mocker.patch("otx.algorithms.action.tasks.train.torch.load", return_value={"state_dict": {}})
        sample_results = {"final_ckpt": "checkpoint_file_path"}
        self.cls_task._get_output_model(sample_results)

    @e2e_pytest_unit
    def test_get_final_eval_results(self, mocker) -> None:
        """Test _get_final_eval_results."""

        class _mock_metric:
            def __init__(self):
                self.performance = Performance(ScoreMetric("accuracy", 1.0))

            def get_performance(self):
                return self.performance

        config = ModelConfiguration(ActionConfig(), self.cls_label_schema)
        output_model = ModelEntity(self.cls_dataset, config)

        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._infer_model", return_value=(True, True))
        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._add_predictions_to_dataset", return_value=True)
        mocker.patch(
            "otx.algorithms.action.tasks.train.ActionTrainTask._add_det_predictions_to_dataset", return_value=True
        )
        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._generate_training_metrics", return_value=[])

        mocker.patch("otx.algorithms.action.tasks.train.ActionTrainTask._get_metric", return_value=_mock_metric())
        self.cls_task._recipe_cfg = Config({"evaluation": {"final_metric": "accuracy"}})
        performance = self.cls_task._get_final_eval_results(self.cls_dataset, output_model)
        assert performance.score.name == "accuracy"
        assert performance.score.value == 1.0

        self.cls_task._task_type = TaskType.ACTION_DETECTION
        performance = self.cls_task._get_final_eval_results(self.cls_dataset, output_model)

    @e2e_pytest_unit
    def test_generate_training_metrics(self) -> None:
        """Test _generate_training_metrics fucntion."""

        class MockCurve:
            def __init__(self, x: list, y: list):
                self.x = x
                self.y = y

        sample_learning_curve = {
            "dummy0": MockCurve([0, 1, 2], [2, 1, 0]),
            "dummy1": MockCurve([0, 1, 2], [2, 1]),
        }
        output = self.cls_task._generate_training_metrics(sample_learning_curve, 1.0, "accuracy")
        assert isinstance(output[0], LineMetricsGroup)
        assert isinstance(output[-1], BarMetricsGroup)



"""Collection of utils for task implementation in Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import math
from typing import List, Optional, Union

from mmcv import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_keys,
    get_configs_by_pairs,
    get_dataset_configs,
    get_meta_keys,
    is_epoch_based_runner,
    patch_color_conversion,
    prepare_work_dir,
    remove_from_config,
    update_config,
)
from otx.algorithms.detection.configs.base import DetectionConfig
from otx.algorithms.detection.utils.data import (
    format_list_to_str,
    get_anchor_boxes,
    get_sizes_from_dataset_entity,
)
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import Domain, LabelEntity
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    DirectoryPathCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

try:
    from sklearn.cluster import KMeans

    __all__ = ["KMeans"]

    KMEANS_IMPORT = True
except ImportError:
    KMEANS_IMPORT = False


logger = get_logger()


@check_input_parameters_type({"work_dir": DirectoryPathCheck})
def patch_config(
    config: Config,
    work_dir: str,
    labels: List[LabelEntity],
):  # pylint: disable=too-many-branches
    """Update config function."""

    # Add training cancelation hook.
    if "custom_hooks" not in config:
        config.custom_hooks = []
    if "CancelTrainingHook" not in {hook.type for hook in config.custom_hooks}:
        config.custom_hooks.append(ConfigDict({"type": "CancelTrainingHook"}))

    # Remove high level data pipelines definition leaving them only inside `data` section.
    remove_from_config(config, "train_pipeline")
    remove_from_config(config, "test_pipeline")

    evaluation_metric = config.evaluation.get("metric")
    if evaluation_metric is not None:
        config.evaluation.save_best = evaluation_metric
    config.checkpoint_config.max_keep_ckpts = 5
    config.checkpoint_config.interval = config.evaluation.get("interval", 1)

    set_data_classes(config, labels)

    config.gpu_ids = range(1)
    config.work_dir = work_dir


@check_input_parameters_type()
def patch_model_config(
    config: Config,
    labels: List[LabelEntity],
):
    """Patch model config."""
    set_num_classes(config, len(labels))


@check_input_parameters_type()
def set_hyperparams(config: Config, hyperparams: DetectionConfig):
    """Set function for hyperparams (DetectionConfig)."""
    config.data.samples_per_gpu = int(hyperparams.learning_parameters.batch_size)
    config.data.workers_per_gpu = int(hyperparams.learning_parameters.num_workers)
    config.optimizer.lr = float(hyperparams.learning_parameters.learning_rate)

    total_iterations = int(hyperparams.learning_parameters.num_iters)

    config.lr_config.warmup_iters = int(hyperparams.learning_parameters.learning_rate_warmup_iters)
    if config.lr_config.warmup_iters == 0:
        config.lr_config.warmup = None
    if is_epoch_based_runner(config.runner):
        config.runner.max_epochs = total_iterations
    else:
        config.runner.max_iters = total_iterations


@check_input_parameters_type()
def patch_adaptive_repeat_dataset(
    config: Union[Config, ConfigDict],
    num_samples: int,
    decay: float = -0.002,
    factor: float = 30,
):
    """Patch the repeat times and training epochs adatively.

    Frequent dataloading inits and evaluation slow down training when the
    sample size is small. Adjusting epoch and dataset repetition based on
    empirical exponential decay improves the training time by applying high
    repeat value to small sample size dataset and low repeat value to large
    sample.

    :param config: mmcv config
    :param num_samples: number of training samples
    :param decay: decaying rate
    :param factor: base repeat factor
    """
    data_train = config.data.train
    if data_train.type == "MultiImageMixDataset":
        data_train = data_train.dataset
    if data_train.type == "RepeatDataset" and getattr(data_train, "adaptive_repeat_times", False):
        if is_epoch_based_runner(config.runner):
            cur_epoch = config.runner.max_epochs
            new_repeat = max(round(math.exp(decay * num_samples) * factor), 1)
            new_epoch = math.ceil(cur_epoch / new_repeat)
            if new_epoch == 1:
                return
            config.runner.max_epochs = new_epoch
            data_train.times = new_repeat


@check_input_parameters_type()
def prepare_for_training(
    config: Union[Config, ConfigDict],
    data_config: ConfigDict,
) -> Union[Config, ConfigDict]:
    """Prepare configs for training phase."""
    prepare_work_dir(config)

    train_num_samples = 0
    for subset in ["train", "val", "test"]:
        data_config_ = data_config.data.get(subset)
        config_ = config.data.get(subset)
        if data_config_ is None:
            continue
        for key in ["otx_dataset", "labels"]:
            found = get_configs_by_keys(data_config_, key, return_path=True)
            if len(found) == 0:
                continue
            assert len(found) == 1
            if subset == "train" and key == "otx_dataset":
                found_value = list(found.values())[0]
                if found_value:
                    train_num_samples = len(found_value)
            update_config(config_, found)

    if train_num_samples > 0:
        patch_adaptive_repeat_dataset(config, train_num_samples)

    return config


@check_input_parameters_type()
def set_data_classes(config: Config, labels: List[LabelEntity]):
    """Setter data classes into config."""
    # Save labels in data configs.
    for subset in ("train", "val", "test"):
        for cfg in get_dataset_configs(config, subset):
            cfg.labels = labels
            #  config.data[subset].labels = labels


@check_input_parameters_type()
def set_num_classes(config: Config, num_classes: int):
    """Set num classes."""
    # Set proper number of classes in model's detection heads.
    head_names = ("mask_head", "bbox_head", "segm_head")
    if "roi_head" in config.model:
        for head_name in head_names:
            if head_name in config.model.roi_head:
                if isinstance(config.model.roi_head[head_name], List):
                    for head in config.model.roi_head[head_name]:
                        head.num_classes = num_classes
                else:
                    config.model.roi_head[head_name].num_classes = num_classes
    else:
        for head_name in head_names:
            if head_name in config.model:
                config.model[head_name].num_classes = num_classes
    # FIXME. ?
    # self.config.model.CLASSES = label_names


@check_input_parameters_type()
def patch_datasets(
    config: Config,
    domain: Domain = Domain.DETECTION,
    subsets: Optional[List[str]] = None,
    **kwargs,
):
    """Update dataset configs."""
    assert "data" in config
    assert "type" in kwargs

    if subsets is None:
        subsets = ["train", "val", "test", "unlabeled"]

    def update_pipeline(cfg):
        if subset == "train":
            for collect_cfg in get_configs_by_pairs(cfg, dict(type="Collect")):
                get_meta_keys(collect_cfg)
        for cfg_ in get_configs_by_pairs(cfg, dict(type="LoadImageFromFile")):
            cfg_.type = "LoadImageFromOTXDataset"
        for cfg_ in get_configs_by_pairs(cfg, dict(type="LoadAnnotations")):
            cfg_.type = "LoadAnnotationFromOTXDataset"
            cfg_.domain = domain
            cfg_.min_size = cfg.pop("min_size", -1)

    for subset in subsets:
        if subset not in config.data:
            continue
        config.data[f"{subset}_dataloader"] = config.data.get(f"{subset}_dataloader", ConfigDict())

        remove_from_config(config.data[subset], "ann_file")
        remove_from_config(config.data[subset], "img_prefix")
        remove_from_config(config.data[subset], "classes")  # Get from DatasetEntity

        cfgs = get_dataset_configs(config, subset)
        for cfg in cfgs:
            cfg.domain = domain
            cfg.otx_dataset = None
            cfg.labels = None
            cfg.update(kwargs)

            remove_from_config(cfg, "ann_file")
            remove_from_config(cfg, "img_prefix")
            remove_from_config(cfg, "classes")  # Get from DatasetEntity

            update_pipeline(cfg)

    patch_color_conversion(config)


def patch_evaluation(config: Config):
    """Update evaluation configs."""
    cfg = config.evaluation
    # CocoDataset.evaluate -> CustomDataset.evaluate
    cfg.pop("classwise", None)
    cfg.metric = "mAP"
    cfg.save_best = "mAP"
    # EarlyStoppingHook
    config.early_stop_metric = "mAP"


def should_cluster_anchors(model_cfg: Config):
    """Check whether cluster anchors or not."""
    if (
        hasattr(model_cfg.model, "bbox_head")
        and hasattr(model_cfg.model.bbox_head, "anchor_generator")
        and getattr(
            model_cfg.model.bbox_head.anchor_generator,
            "reclustering_anchors",
            False,
        )
    ):
        return True
    return False


@check_input_parameters_type({"dataset": DatasetParamTypeCheck})
def cluster_anchors(model_config: Config, data_config: Config, dataset: DatasetEntity):
    """Update configs for cluster_anchors."""
    if not KMEANS_IMPORT:
        raise ImportError(
            "Sklearn package is not installed. To enable anchor boxes clustering, please install "
            "packages from requirements/optional.txt or just scikit-learn package."
        )

    logger.info("Collecting statistics from training dataset to cluster anchor boxes...")
    [target_wh] = [
        transforms.img_scale for transforms in data_config.data.test.pipeline if transforms.type == "MultiScaleFlipAug"
    ]
    prev_generator = model_config.model.bbox_head.anchor_generator
    group_as = [len(width) for width in prev_generator.widths]
    wh_stats = get_sizes_from_dataset_entity(dataset, list(target_wh))

    if len(wh_stats) < sum(group_as):
        logger.warning(
            f"There are not enough objects to cluster: {len(wh_stats)} were detected, while it should be "
            f"at least {sum(group_as)}. Anchor box clustering was skipped."
        )
        return

    widths, heights = get_anchor_boxes(wh_stats, group_as)
    logger.info(
        f"Anchor boxes widths have been updated from {format_list_to_str(prev_generator.widths)} "
        f"to {format_list_to_str(widths)}"
    )
    logger.info(
        f"Anchor boxes heights have been updated from {format_list_to_str(prev_generator.heights)} "
        f"to {format_list_to_str(heights)}"
    )
    config_generator = model_config.model.bbox_head.anchor_generator
    config_generator.widths, config_generator.heights = widths, heights

    model_config.model.bbox_head.anchor_generator = config_generator



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Callable, Dict, List, Optional, Tuple, Union

import openvino.runtime as ov

from otx.mpa.utils.logger import get_logger

from ..graph.parsers.builder import PARSERS
from .ov_model import OVModel

logger = get_logger()


class ParserMixin:
    def parse(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        parser: Optional[Union[str, Callable]] = None,
        **kwargs,
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:

        parser = self.parser if parser is None else parser
        if isinstance(parser, str):
            parser = PARSERS.get(parser)

        if not inputs or not outputs:
            graph = OVModel.build_graph(model_path_or_model, weight_path)
            parsed = parser(graph, **kwargs)

            if not isinstance(parsed, dict) or ("inputs" not in parsed and "outputs" not in parsed):
                raise ValueError(f"parser {parser} failed to find inputs and outputs of model. ")
            if isinstance(parsed["inputs"], dict) != isinstance(parsed["outputs"], dict):
                raise ValueError(f"output of parser ({parser}) is not consistent")
            if isinstance(parsed["inputs"], dict) and isinstance(parsed["outputs"], dict):
                if set(parsed["inputs"].keys()) != set(parsed["outputs"].keys()):
                    raise ValueError(
                        f"input keys {parsed['inputs'].keys()} and "
                        f"output keys {parsed['outputs'].keys()} are different."
                    )

            inputs = parsed["inputs"] if not inputs else inputs
            outputs = parsed["outputs"] if not outputs else outputs
            logger.info(f"inputs: {inputs}")
            logger.info(f"outputs: {outputs}")

        return inputs, outputs

    @staticmethod
    def parser(graph, **kwargs) -> Dict[str, Union[List[str], Dict[str, List[str]]]]:
        return dict(inputs=[], outputs=[])



"""Progressbar  and Score Reporting callback Callback for OTX task.

TODO Since only one progressbar callback is supported HPO is combined into one callback. Remove this after the refactor
"""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Optional, Union

from pytorch_lightning.callbacks.progress import TQDMProgressBar

from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.train_parameters import TrainParameters, default_progress_callback


class ProgressCallback(TQDMProgressBar):
    """Progress Callback.

    Modify progress callback to show completion of the entire training step.
    """

    def __init__(
        self, parameters: Optional[Union[TrainParameters, InferenceParameters, OptimizationParameters]] = None
    ) -> None:
        super().__init__()
        self.current_epoch: int = 0
        self.max_epochs: int = 0
        self._progress: float = 0

        if parameters is not None:
            self.progress_and_hpo_callback = parameters.update_progress
        else:
            self.progress_and_hpo_callback = default_progress_callback

    def on_train_start(self, trainer, pl_module):
        """Store max epochs and current epoch from trainer."""
        super().on_train_start(trainer, pl_module)
        self.current_epoch = trainer.current_epoch
        self.max_epochs = trainer.max_epochs
        self._reset_progress()

    def on_predict_start(self, trainer, pl_module):
        """Reset progress bar when prediction starts."""
        super().on_predict_start(trainer, pl_module)
        self._reset_progress()

    def on_test_start(self, trainer, pl_module):
        """Reset progress bar when testing starts."""
        super().on_test_start(trainer, pl_module)
        self._reset_progress()

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        """Adds training completion percentage to the progress bar."""
        super().on_train_batch_end(trainer, pl_module, outputs, batch, batch_idx)
        self.current_epoch = trainer.current_epoch
        self._update_progress(stage="train")

    def on_predict_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        """Adds prediction completion percentage to the progress bar."""
        super().on_predict_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)
        self._update_progress(stage="predict")

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        """Adds testing completion percentage to the progress bar."""
        super().on_test_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)
        self._update_progress(stage="test")

    def on_validation_epoch_end(self, trainer, pl_module):  # pylint: disable=unused-argument
        """If score exists in trainer.logged_metrics, report the score."""
        if self.progress_and_hpo_callback is not None:
            score = None
            metric = getattr(self.progress_and_hpo_callback, "metric", None)
            if metric in trainer.logged_metrics:
                score = float(trainer.logged_metrics[metric])

            # Always assumes that hpo validation step is called during training.
            self.progress_and_hpo_callback(int(self._get_progress("train")), score)  # pylint: disable=not-callable

    def _reset_progress(self):
        self._progress = 0.0

    def _get_progress(self, stage: str = "train") -> float:
        """Get progress for train and test stages.

        Args:
            stage (str, optional): Train or Test stages. Defaults to "train".
        """
        if stage == "train":
            # Progress is calculated on the upper bound (max epoch).
            # Early stopping might stop the training before the progress reaches 100%
            self._progress = (
                (self.train_batch_idx + self.current_epoch * self.total_train_batches)
                / (self.total_train_batches * self.max_epochs)
            ) * 100

        elif stage == "predict":
            self._progress = (self.predict_batch_idx / (self.total_predict_batches_current_dataloader + 1e-10)) * 100

        elif stage == "test":
            self._progress = (self.test_batch_idx / (self.total_test_batches_current_dataloader + 1e-10)) * 100
        else:
            raise ValueError(f"Unknown stage {stage}. Available: train, predict and test")

        return self._progress

    def _update_progress(self, stage: str):
        progress = self._get_progress(stage)
        self.progress_and_hpo_callback(int(progress), None)



"""Tests the inference callback on a dummy lightning module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import pytest
import pytorch_lightning as pl

from otx.algorithms.anomaly.adapters.anomalib.callbacks import AnomalyInferenceCallback
from otx.api.entities.model_template import TaskType
from tests.unit.algorithms.anomaly.helpers.dummy_dataset import DummyDataModule
from tests.unit.algorithms.anomaly.helpers.dummy_model import DummyModel


class TestInferenceCallback:
    @pytest.mark.parametrize(
        "task_type", [TaskType.ANOMALY_CLASSIFICATION, TaskType.ANOMALY_DETECTION, TaskType.ANOMALY_SEGMENTATION]
    )
    def test_inference_callback(self, task_type):
        """For each task type test the inference callback.

        The inference callback is responsible for processing the predictions and generating the annotations.

        Args:
            task_type (TaskType): Task type.
        """
        datamodule = DummyDataModule(task_type)
        model = DummyModel()
        labels = datamodule.labels
        callback = AnomalyInferenceCallback(datamodule.dataset.dataset, labels, task_type)
        trainer = pl.Trainer(logger=False, callbacks=[callback])
        result = trainer.predict(model, datamodule=datamodule)
        # TODO: Currently it only checks that the result has predicted labels. This should be expanded to check the
        # box labels and masks based on the task type.
        assert result[0]["pred_labels"].item() == 1.0



"""Tiling for detection and instance segmentation task."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import os.path as osp
import tempfile
import uuid
from itertools import product
from multiprocessing import Pool
from time import time
from typing import Callable, Dict, List, Tuple, Union

import mmcv
import numpy as np
import pycocotools.mask as mask_util
from mmcv.ops import nms
from mmdet.core import BitmapMasks, bbox2result
from tqdm import tqdm


def timeit(func) -> Callable:
    """Decorator to measure time of function execution.

    Args:
        func:

    Returns:
        Callable function with time measurement.
    """

    def wrapper(*args, **kwargs):
        begin = time()
        result = func(*args, **kwargs)
        print(f"\n==== {func.__name__}: {time() - begin} sec ====\n")
        return result

    return wrapper


# pylint: disable=too-many-instance-attributes, too-many-arguments
class Tile:
    """Tile and merge datasets.

    Args:
        dataset (CustomDataset): the dataset to be tiled.
        tile_size (int): the length of side of each tile. Defaults to 400
        overlap (float, optional): ratio of each tile to overlap with each of
            the tiles in its 4-neighborhood. Defaults to 0.2.
        min_area_ratio (float, optional): the minimum overlap area ratio
            between a tiled image and its annotations. Ground-truth box is
            discarded if the overlap area is less than this value.
            Defaults to 0.9.
        iou_threshold (float, optional): IoU threshold to be used to suppress
            boxes in tiles' overlap areas. Defaults to 0.45.
        max_per_img (int, optional): if there are more than max_per_img bboxes
            after NMS, only top max_per_img will be kept. Defaults to 1500.
        filter_empty_gt (bool, optional): If set true, images without bounding
            boxes of the dataset's classes will be filtered out. This option
            only works when `test_mode=False`, i.e., we never filter images
            during tests. Defaults to True.
        nproc (int, optional): Processes used for processing masks. Default: 4.
    """

    def __init__(
        self,
        dataset,
        pipeline,
        tmp_dir: tempfile.TemporaryDirectory,
        tile_size: int = 400,
        overlap: float = 0.2,
        min_area_ratio: float = 0.9,
        iou_threshold: float = 0.45,
        max_per_img: int = 1500,
        filter_empty_gt: bool = True,
        nproc: int = 4,
    ):
        self.min_area_ratio = min_area_ratio
        self.filter_empty_gt = filter_empty_gt
        self.iou_threshold = iou_threshold
        self.max_per_img = max_per_img
        self.tile_size = tile_size
        self.overlap = overlap
        self.stride = int(tile_size * (1 - overlap))
        self.num_images = len(dataset)
        self.num_classes = len(dataset.CLASSES)
        self.CLASSES = dataset.CLASSES  # pylint: disable=invalid-name
        self.tmp_folder = tmp_dir.name
        self.nproc = nproc
        self.img2fp32 = False
        for p in pipeline:
            if p.type == "PhotoMetricDistortion":
                self.img2fp32 = True
                break

        self.dataset = dataset
        self.tiles = self.gen_tile_ann()
        self.cache_tiles()

    @timeit
    def cache_tiles(self):
        """Cache tiles to disk."""
        pbar = tqdm(total=len(self.tiles))
        pre_img_idx = None
        for i, tile in enumerate(self.tiles):
            tile["tile_path"] = osp.join(
                self.tmp_folder, "_".join([str(i), tile["uuid"], tile["ori_filename"], ".jpg"])
            )
            x_1, y_1, x_2, y_2 = tile["tile_box"]
            dataset_idx = tile["dataset_idx"]
            if dataset_idx != pre_img_idx:
                ori_img = self.dataset[dataset_idx]["img"]
                pre_img_idx = dataset_idx

            mmcv.imwrite(ori_img[y_1:y_2, x_1:x_2, :], tile["tile_path"])
            pbar.update(1)

    @timeit
    def gen_tile_ann(self) -> List[Dict]:
        """Generate tile information and tile annotation from dataset.

        Returns:
            List[Dict]: A list of tiles generated from the dataset. Each item comprises tile annotation and tile
                        coordinates relative to the original image.
        """
        tiles = []
        pbar = tqdm(total=len(self.dataset))

        for idx, result in enumerate(self.dataset):
            tiles.append(self.gen_single_img(result, dataset_idx=idx))

        for idx, result in enumerate(self.dataset):
            tiles.extend(self.gen_tiles_single_img(result, dataset_idx=idx))
            pbar.update(1)
        return tiles

    def gen_single_img(self, result: Dict, dataset_idx: int) -> Dict:
        """Add full-size image for inference or training.

        Args:
            result (Dict): the original image-level result (i.e. the original image annotation)
            dataset_idx (int): the image index this tile belongs to

        Returns:
            Dict: annotation with some other useful information for data pipeline.
        """
        result["tile_box"] = (0, 0, result["dataset_item"].width, result["dataset_item"].height)
        result["dataset_idx"] = dataset_idx
        result["original_shape_"] = result["img_shape"]
        result["uuid"] = str(uuid.uuid4())
        return result

    # pylint: disable=too-many-locals
    def gen_tiles_single_img(self, result: Dict, dataset_idx: int) -> List[Dict]:
        """Generate tile annotation for a single image.

        Args:
            result (Dict): the original image-level result (i.e. the original image annotation)
            dataset_idx (int): the image index this tile belongs to

        Returns:
            List[Dict]: a list of tile annotation with some other useful information for data pipeline.
        """
        tile_list = []
        gt_bboxes = result.pop("gt_bboxes", np.zeros((0, 4), dtype=np.float32))
        gt_masks = result.pop("gt_masks", None)
        gt_bboxes_ignore = result.pop("gt_bboxes_ignore", np.zeros((0, 4), dtype=np.float32))
        gt_labels = result.pop("gt_labels", np.array([], dtype=np.int64))
        img_shape = result.pop("img_shape")
        height, width = img_shape[:2]
        _tile = self.prepare_result(result)

        num_patches_h = int((height - self.tile_size) / self.stride) + 1
        num_patches_w = int((width - self.tile_size) / self.stride) + 1
        for (_, _), (loc_i, loc_j) in zip(
            product(range(num_patches_h), range(num_patches_w)),
            product(
                range(0, height - self.tile_size + 1, self.stride),
                range(0, width - self.tile_size + 1, self.stride),
            ),
        ):
            x_1 = loc_j
            x_2 = loc_j + self.tile_size
            y_1 = loc_i
            y_2 = loc_i + self.tile_size
            tile = copy.deepcopy(_tile)
            tile["original_shape_"] = img_shape
            tile["ori_shape"] = (y_2 - y_1, x_2 - x_1, 3)
            tile["img_shape"] = tile["ori_shape"]
            tile["tile_box"] = (x_1, y_1, x_2, y_2)
            tile["dataset_idx"] = dataset_idx
            tile["gt_bboxes_ignore"] = gt_bboxes_ignore
            tile["uuid"] = str(uuid.uuid4())
            self.tile_ann_assignment(tile, np.array([[x_1, y_1, x_2, y_2]]), gt_bboxes, gt_masks, gt_labels)
            # filter empty ground truth
            if self.filter_empty_gt and len(tile["gt_labels"]) == 0:
                continue
            tile_list.append(tile)
        return tile_list

    def prepare_result(self, result: Dict) -> Dict:
        """Prepare results dict for pipeline.

        Args:
            result (Dict): original image-level result for a tile

        Returns:
            Dict: result template with useful information for data pipeline.
        """
        result_template = dict(
            ori_filename=result["ori_filename"],
            filename=result["filename"],
            bbox_fields=result["bbox_fields"],
            mask_fields=result["mask_fields"],
            seg_fields=result["seg_fields"],
            img_fields=result["img_fields"],
        )
        return result_template

    def tile_ann_assignment(
        self,
        tile_result: Dict,
        tile_box: np.ndarray,
        gt_bboxes: np.ndarray,
        gt_masks: BitmapMasks,
        gt_labels: np.ndarray,
    ):
        """Assign new annotation to this tile.

        Ground-truth is discarded if the overlap with this tile is lower than
        min_area_ratio.

        Args:
            tile_result (Dict): the tile-level result (i.e. the tile annotation)
            tile_box (np.ndarray): the coordinate for this tile box (i.e. the tile coordinate relative to the image)
            gt_bboxes (np.ndarray): the original image-level boxes
            gt_masks (BitmapMasks): the original image-level masks
            gt_labels (np.ndarray): the original image-level labels
        """
        x_1, y_1 = tile_box[0][:2]
        overlap_ratio = self.tile_boxes_overlap(tile_box, gt_bboxes)
        match_idx = np.where((overlap_ratio[0] >= self.min_area_ratio))[0]

        if len(match_idx):
            tile_lables = gt_labels[match_idx][:]
            tile_bboxes = gt_bboxes[match_idx][:]
            tile_bboxes[:, 0] -= x_1
            tile_bboxes[:, 1] -= y_1
            tile_bboxes[:, 2] -= x_1
            tile_bboxes[:, 3] -= y_1
            tile_bboxes[:, 0] = np.maximum(0, tile_bboxes[:, 0])
            tile_bboxes[:, 1] = np.maximum(0, tile_bboxes[:, 1])
            tile_bboxes[:, 2] = np.minimum(self.tile_size, tile_bboxes[:, 2])
            tile_bboxes[:, 3] = np.minimum(self.tile_size, tile_bboxes[:, 3])
            tile_result["gt_bboxes"] = tile_bboxes
            tile_result["gt_labels"] = tile_lables
            tile_result["gt_masks"] = gt_masks[match_idx].crop(tile_box[0]) if gt_masks is not None else []
        else:
            tile_result.pop("bbox_fields")
            tile_result.pop("mask_fields")
            tile_result.pop("seg_fields")
            tile_result.pop("img_fields")
            tile_result["gt_bboxes"] = []
            tile_result["gt_labels"] = []
            tile_result["gt_masks"] = []

        if gt_masks is None:
            tile_result.pop("gt_masks")

    def tile_boxes_overlap(self, tile_box: np.ndarray, boxes: np.ndarray) -> np.ndarray:
        """Compute overlapping ratio over boxes.

        Args:
            tile_box (np.ndarray): box in shape (1, 4).
            boxes (np.ndarray): boxes in shape (N, 4).

        Returns:
            np.ndarray: overlapping ratio over boxes
        """
        box_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])

        width_height = np.minimum(tile_box[:, None, 2:], boxes[:, 2:]) - np.maximum(tile_box[:, None, :2], boxes[:, :2])

        width_height = width_height.clip(min=0)  # [N,M,2]
        inter = width_height.prod(2)

        # handle empty boxes
        tile_box_ratio = np.where(inter > 0, inter / box_area, np.zeros(1, dtype=inter.dtype))
        return tile_box_ratio

    def multiclass_nms(
        self, boxes: np.ndarray, scores: np.ndarray, idxs: np.ndarray, iou_threshold: float, max_num: int
    ):
        """NMS for multi-class bboxes.

        Args:
            boxes (np.ndarray):  boxes in shape (N, 4).
            scores (np.ndarray): scores in shape (N, ).
            idxs (np.ndarray):  each index value correspond to a bbox cluster,
                and NMS will not be applied between elements of different idxs,
                shape (N, ).
            iou_threshold (float): IoU threshold to be used to suppress boxes
                in tiles' overlap areas.
            max_num (int): if there are more than max_per_img bboxes after
                NMS, only top max_per_img will be kept.

        Returns:
            tuple: tuple: kept dets and indice.
        """
        if len(boxes) == 0:
            return None, []
        max_coordinate = boxes.max()
        offsets = idxs.astype(boxes.dtype) * (max_coordinate + 1)
        boxes_for_nms = boxes + offsets[:, None]
        dets, keep = nms(boxes_for_nms, scores, iou_threshold)
        if max_num > 0:
            dets = dets[:max_num]
            keep = keep[:max_num]
        return dets, keep

    def tile_nms(
        self,
        bbox_results: List[np.ndarray],
        mask_results: List[List],
        label_results: List[np.ndarray],
        iou_threshold: float,
        max_per_img: int,
        detection: bool,
    ):
        """NMS after aggregation suppressing duplicate boxes in tile-overlap areas.

        Args:
            bbox_results (List[List]): image-level box prediction
            mask_results (List[np.ndarray]): image-level mask prediction
            label_results (List[List]): image-level label prediction
            iou_threshold (float): IoU threshold to be used to suppress boxes in tiles' overlap areas.
            max_per_img (int): if there are more than max_per_img bboxes after NMS, only top max_per_img will be kept.
            detection (bool): whether it is a detection task
        """
        assert len(bbox_results) == len(mask_results) == len(label_results)
        for i, result in enumerate(zip(bbox_results, mask_results, label_results)):
            score_bboxes, masks, labels = result
            bboxes = score_bboxes[:, :4]
            scores = np.ascontiguousarray(score_bboxes[:, 4])
            _, keep_indices = self.multiclass_nms(
                bboxes, scores, labels, iou_threshold=iou_threshold, max_num=max_per_img
            )

            bboxes = bboxes[keep_indices]
            labels = labels[keep_indices]
            scores = scores[keep_indices]
            bbox_results[i] = bbox2result(np.concatenate([bboxes, scores[:, None]], -1), labels, self.num_classes)

            if not detection:
                masks = [masks[keep_idx] for keep_idx in keep_indices]
                masks = self.process_masks(masks)
                mask_results[i] = [list(np.asarray(masks)[labels == i]) for i in range(self.num_classes)]

    def __len__(self):
        """Total number of tiles."""
        return len(self.tiles)

    def __getitem__(self, idx):
        """Get training/test tile.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training/test data.
        """
        result = copy.deepcopy(self.tiles[idx])
        if result.get("tile_path") and osp.isfile(result["tile_path"]):
            img = mmcv.imread(result["tile_path"])
            if self.img2fp32:
                img = img.astype(np.float32)
            result["img"] = img
            return result
        dataset_idx = result["dataset_idx"]
        x_1, y_1, x_2, y_2 = result["tile_box"]
        ori_img = self.dataset[dataset_idx]["img"]
        if self.img2fp32:
            ori_img = ori_img.astype(np.float32)
        result["img"] = ori_img[y_1:y_2, x_1:x_2, :]
        return result

    @staticmethod
    def readjust_tile_mask(tile_rle: Dict):
        """Shift tile-level mask to image-level mask.

        Args:
            tile_rle (Dict): _description_

        Returns:
            _type_: _description_
        """
        x1, y1, x2, y2 = tile_rle.pop("tile_box")
        height, width = tile_rle.pop("img_size")
        tile_mask = mask_util.decode(tile_rle)
        tile_mask = np.pad(tile_mask, ((y1, height - y2), (x1, width - x2)))
        return mask_util.encode(tile_mask)

    def process_masks(self, tile_masks: List[Dict]):
        """Decode Mask Result to Numpy mask, add paddings then encode masks again.

        Args:
            tile_masks (_type_): _description_

        Returns:
            _type_: _description_
        """
        with Pool(self.nproc) as pool:
            results = pool.map(Tile.readjust_tile_mask, tile_masks)
        return results

    # pylint: disable=too-many-locals
    @timeit
    def merge(self, results: List[List]) -> Union[List[Tuple[np.ndarray, list]], List[np.ndarray]]:
        """Merge/Aggregate tile-level prediction to image-level prediction.

        Args:
            results (list[list | tuple]): Testing tile results of the dataset.

        Returns:
            List[List]: Testing image results of the dataset.
        """
        assert len(results) == len(self.tiles)

        detection = False
        if isinstance(results[0], tuple):
            num_classes = len(results[0][0])
            dtype = results[0][0][0].dtype
        elif isinstance(results[0], list):
            detection = True
            num_classes = len(results[0])
            dtype = results[0][0].dtype
        else:
            raise RuntimeError("Unknown data type")

        merged_bbox_results: List[np.ndarray] = [np.empty((0, 5), dtype=dtype) for _ in range(self.num_images)]
        merged_mask_results: List[List] = [[] for _ in range(self.num_images)]
        merged_label_results: List[Union[List, np.ndarray]] = [[] for _ in range(self.num_images)]

        for result, tile in zip(results, self.tiles):
            tile_x1, tile_y1, _, _ = tile["tile_box"]
            img_idx = tile["dataset_idx"]
            img_h, img_w, _ = tile["original_shape_"]

            mask_result: List[List] = [[] for _ in range(num_classes)]
            if isinstance(result, tuple):
                bbox_result, mask_result = result
            else:
                bbox_result = result

            for cls_idx, cls_result in enumerate(zip(bbox_result, mask_result)):
                cls_bbox_result, cls_mask_result = cls_result
                _tmp_cls_bbox_result = np.zeros_like(cls_bbox_result)
                _tmp_cls_bbox_result[:, 0] = cls_bbox_result[:, 0] + tile_x1
                _tmp_cls_bbox_result[:, 1] = cls_bbox_result[:, 1] + tile_y1
                _tmp_cls_bbox_result[:, 2] = cls_bbox_result[:, 2] + tile_x1
                _tmp_cls_bbox_result[:, 3] = cls_bbox_result[:, 3] + tile_y1
                _tmp_cls_bbox_result[:, 4] = cls_bbox_result[:, 4]

                merged_bbox_results[img_idx] = np.concatenate((merged_bbox_results[img_idx], _tmp_cls_bbox_result))
                merged_label_results[img_idx] = np.concatenate(
                    [merged_label_results[img_idx], len(cls_bbox_result) * [cls_idx]]
                )

                for cls_mask_dict in cls_mask_result:
                    cls_mask_dict.update(dict(tile_box=tile["tile_box"], img_size=(img_h, img_w)))
                merged_mask_results[img_idx] += cls_mask_result

        # run NMS after aggregation suppressing duplicate boxes in
        # overlapping areas
        self.tile_nms(
            merged_bbox_results,
            merged_mask_results,
            merged_label_results,
            iou_threshold=self.iou_threshold,
            max_per_img=self.max_per_img,
            detection=detection,
        )

        assert len(merged_bbox_results) == len(merged_mask_results)
        if detection:
            return list(merged_bbox_results)
        return list(zip(merged_bbox_results, merged_mask_results))



"""Interface for inferencer."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
import logging
import multiprocessing
import queue
import warnings
from pathlib import Path
from typing import Any, Iterator, List, Optional, Tuple, Union

import numpy as np

# pylint: disable=no-name-in-module
from openvino.inference_engine import ExecutableNetwork, IECore, InferRequest
from openvino.inference_engine.constants import OK, RESULT_NOT_READY

from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.usecases.exportable_code.streamer.streamer import BaseStreamer

__all__ = [
    "AsyncOpenVINOTask",
    "BaseInferencer",
    "BaseOpenVINOInferencer",
    "IInferencer",
]

logger = logging.getLogger(__name__)


class IInferencer(metaclass=abc.ABCMeta):
    """Base interface class for the inference task.

    This class could be used by both the analyse method in the task, and the exportable code inference.

    """

    @abc.abstractmethod
    def pre_process(self, image: np.ndarray) -> Tuple[Any, Any]:
        """Pre-process input image and return the pre-processed image with meta-data if required.

        This method should pre-process the input image, and return the processed output and if required a Tuple with
        metadata that is required for post_process to work.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def forward(self, image: Any) -> Any:
        """Forward the input image to the model and return the output.

        NOTE: The input is typed as Any at the moment, mainly because it could be numpy
            array,torch Tensor or tf Tensor. In the future, it could be an idea to be
            more specific.

        This method should perform the prediction by forward-passing the input image
        to the model, and return the predictions in a dictionary format.

        For instance, for a segmentation task, the predictions could be {"mask": mask}.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def post_process(self, prediction: Any, metadata: Any) -> AnnotationSceneEntity:
        """Post-process the raw predictions, and return the AnnotationSceneEntity.

        This method should include the post-processing methods that are applied to the raw predictions from the
        self.forward() stage.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, image: np.ndarray) -> AnnotationSceneEntity:
        """This method performs a prediction."""
        raise NotImplementedError


class BaseInferencer(IInferencer, abc.ABC):
    """Base class for standard inference.

    The user needs to implement the following:
        + `load_model`
        + `pre_process`
        + `forward`
        + `post_process`
    """

    def predict(self, image: np.ndarray) -> AnnotationSceneEntity:
        """Perform a prediction for a given input image.

        Args:
            image: Input image

        Returns:
            Output predictions
        """
        image, metadata = self.pre_process(image)
        predictions = self.forward(image)
        predictions = self.post_process(predictions, metadata)

        return predictions


class BaseOpenVINOInferencer(BaseInferencer, abc.ABC):
    """Base class for OpenVINO inference.

    Can handle the basic flow of reading and loading a model. If the network needs to be reshaped,
    override the load_model function.
    One would need to implement the following methods to use this as OpenVINO
    Inferencer
        + `pre_process`
        + `forward`
        + `post_process`

    Args:
        weight_path: Path to the weight file
        device: Device to use for inference. Check available devices
            with IECore().available_devices.
        num_requests: Number of simultaneous requests that can be issued
            to the model, has no effect on synchronous execution.

    Raises:
        ValueError: Raised if the device is not available.
    """

    def __init__(
        self,
        model_file: Union[str, bytes],
        weights_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):
        self.ie_core = IECore()
        if device not in self.ie_core.available_devices:
            raise ValueError(
                f"Device '{device}' is not available for inference. Available devices "
                f"are: {self.ie_core.available_devices}"
            )

        self.device: str = device
        self.num_requests: int = num_requests

        self.input_keys: Optional[List[str]] = None
        self.output_keys: Optional[List[str]] = None

        self.net: Optional[ExecutableNetwork] = None
        self.model: ExecutableNetwork
        self.load_model(model_file, weights_file)

    def read_model(
        self,
        model_file: Union[Path, str, bytes],
        weights_file: Union[Path, str, bytes, None] = None,
    ):
        """Reads an OpenVINO model and saves its input and output keys to a list.

        Args:
            model_file: Path to the model file or bytes with data from OpenVINO's .xml file.
            weights: A .xml, .bin or .onnx file to be loaded. if a .xml
                or .bin is provided a file of the same name with the
                other extension is also expected

        Raises:
            ValueError: Raised if a weights file that is not compatible with OpenVINO
                        is provided
        """
        if isinstance(model_file, str):
            path = Path(model_file)
            if path.suffix == ".onnx":
                weights_file = None
            elif path.suffix in (".xml", ".bin"):
                model_file = path.with_suffix(".xml")
                weights_file = path.with_suffix(".bin")
            else:
                raise ValueError(f"Unsupported file extension: {path.suffix}")

        init_from_buffer = isinstance(model_file, bytes)
        self.net = self.ie_core.read_network(model=model_file, weights=weights_file, init_from_buffer=init_from_buffer)
        self.input_keys = list(self.net.input_info.keys())
        self.output_keys = list(self.net.outputs.keys())

    def load_model(self, model_file: Union[str, bytes], weights_file: Union[str, bytes, None]):
        """Loads an OpenVINO or ONNX model, overwrite this function if you need to reshape the network.

        Or retrieve additional information from the network after loading it.

        Args:
            model_file (Union[str, bytes]): Path to the model file or bytes with data from OpenVINO's .xml file.
            weights_file (weights_file: Union[str, bytes, None]): A .xml, .bin or .onnx file to be loaded. if a .xml
                or .bin is provided a file of the same name with the
                other extension is also expected
        """
        if self.net is None:
            self.read_model(model_file, weights_file)

        self.model = self.ie_core.load_network(
            network=self.net, device_name=self.device, num_requests=self.num_requests
        )


class AsyncOpenVINOTask:
    """This class runs asynchronous inference on a BaseOpenVinoInferencer.

    Using a BaseStreamer as input

    Args:
        streamer: A streamer that provides input for the inferencer
        inferencer: The inferencer to use to generate predictions
        drop_output: Set to a number to limit the amount of results
            stored at a time. If inference is completed but there is
            no room for the output. The output will be dropped.Set
            to 0 to disable, Set to None to automatically determine
            a good value
    """

    def __init__(
        self,
        streamer: BaseStreamer,
        inferencer: BaseOpenVINOInferencer,
        drop_output: Optional[int] = None,
    ):
        self.streamer: BaseStreamer = streamer
        self.inferencer: BaseOpenVINOInferencer = inferencer

        if drop_output is None:
            # Setting to 2x number of requests should allow for a good balance
            # between memory conservation and flexibility.
            drop_output = self.inferencer.num_requests * 2

        self.drop_output = drop_output

    def __iter__(self) -> Iterator[Tuple[np.ndarray, List[np.ndarray]]]:
        """Starts the asynchronous inference loop.

        Example:
            >>> streamer = VideoStreamer("../demo.mp4")
            >>> inferencer = ExampleOpenVINOInferencer(weights="model.bin", num_requests=4)
            >>> async_task = AsyncOpenVINOTask(streamer, inferencer)
            >>> for image, predictions in async_task:
            ...    # Do something with predictions

        Yields:
            Iterator[Tuple[np.ndarray, List[np.ndarray]]]: A Tuple with the used image and a list of predictions
        """
        manager = multiprocessing.Manager()
        completed_requests = manager.Queue(maxsize=self.drop_output)

        if len(self.inferencer.model.requests) == 1:
            warnings.warn(
                "Using AsyncOpenVINOTask while num_requests of the inferencer is set "
                "to 1, use a higher value to get a performance benefit or use "
                "synchronous execution."
            )

        try:
            for frame in self.streamer:
                while not self.__idle_request_available():
                    try:
                        yield completed_requests.get(timeout=0.1)
                    except queue.Empty:
                        pass
                self.__make_request(frame, completed_requests)

            while not completed_requests.empty():
                try:
                    yield completed_requests.get(timeout=0.1)
                except queue.Empty:
                    self.__wait_for_request()
        except GeneratorExit:
            pass

    def __idle_request_available(self) -> bool:
        """Returns True if one idle request is available.

        Returns:
            bool: True if one idle request is available
        """
        return self.inferencer.model.get_idle_request_id() >= 0

    def __wait_for_request(self, num_requests: Optional[int] = None, timeout: Optional[int] = None) -> bool:
        """Wait for num_requests to become available.

        Args:
            num_requests: Number of requests that should be available
                for the function to return False. If set to None waits
                for all requests to finish. Defaults to None.
            timeout: Amount of milliseconds to wait before function
                returns regardless of available requests. Set to None to
                wait regardless of the time. Defaults to None.

        Returns:
            bool -- Returns True if no requests are available, False if
            num_requests are available
        """
        return self.inferencer.model.wait(num_requests=num_requests, timeout=timeout) == RESULT_NOT_READY

    def __make_request(self, image: np.ndarray, completed_requests: queue.Queue):
        """Makes an asynchronous request.

        Args:
            image: Image to run inference on.
            completed_requests: Queue where results should be placed.

        Raises:
            RuntimeError: Raised if no idle requests are available
        """
        request_id = self.inferencer.model.get_idle_request_id()
        if request_id == -1:
            raise RuntimeError("Tried to get idle request but got no request")
        request = self.inferencer.model.requests[request_id]
        input_image, metadata = self.inferencer.pre_process(image)

        request.set_completion_callback(
            py_callback=_async_callback,
            py_data=(self, request, image, metadata, completed_requests),
        )
        request.async_infer(inputs=input_image)


def _async_callback(
    status,
    callback_args: Tuple[AsyncOpenVINOTask, InferRequest, np.ndarray, Any, multiprocessing.Queue],
):
    """Callback for Async Infer.

    Adds the used image and output Dictionary to the completed_requests Queue.

    Args:
        status: OpenVINO status code
        callback_args: AsyncOpenVINOTask object, the Inference Request,
            the image used, Queue to put the output in.
    """
    self, request, image, metadata, completed_requests = callback_args
    try:
        if status != OK:
            raise RuntimeError(f"Infer request has returned status code {status}")

        output_blobs = request.output_blobs
        output_blobs = {k: output_blob.buffer for (k, output_blob) in output_blobs.items()}
        output = self.inferencer.post_process(output_blobs, metadata)

        try:
            completed_requests.put_nowait((image, output))
        except queue.Full:
            logger.warning("An inference result was dropped because the queue is full")
            # TODO: Make the callback safely drop the oldest output and add the
            #  new output
        except AttributeError:
            # If __iter__ has exited while requests are still up this exception is
            # thrown
            pass

    except RuntimeError as error:
        logger.warning("RunTimeError in AsyncOpenVINOTask: _async_callback: %s", str(error))



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.seg.stage import SegStage
from otx.mpa.utils.config_utils import remove_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class SemiSLSegStage(SegStage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def configure_data(self, cfg, training, data_cfg, **kwargs):
        """Patch cfg.data."""
        super().configure_data(cfg, training, data_cfg, **kwargs)
        # Set unlabeled data hook
        if training:
            if cfg.data.get("unlabeled", False) and cfg.data.unlabeled.get("otx_dataset", False):
                self.configure_unlabeled_dataloader(cfg, self.distributed)

    def configure_task(self, cfg, training, **kwargs):
        """Adjust settings for task adaptation"""
        super().configure_task(cfg, training, **kwargs)

        # Don't pass task_adapt arg to semi-segmentor
        if cfg.model.type != "ClassIncrEncoderDecoder" and cfg.model.get("task_adapt", False):
            cfg.model.pop("task_adapt")

        # Remove task adapt hook (set default torch random sampler)
        remove_custom_hook(cfg, "TaskAdaptHook")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.det.trainer import DetectionTrainer
from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import IncrDetectionStage

logger = get_logger()


@STAGES.register_module()
class IncrDetectionTrainer(IncrDetectionStage, DetectionTrainer):
    def __init__(self, **kwargs):
        IncrDetectionStage.__init__(self, **kwargs)



"""Openvino Task of Detection."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import io
import json
import multiprocessing
import os
import tempfile
import warnings
from typing import Any, Dict, Optional, Tuple, Union
from zipfile import ZipFile

import attr
import numpy as np
from addict import Dict as ADDict
from compression.api import DataLoader
from compression.engines.ie_engine import IEEngine
from compression.graph import load_model, save_model
from compression.graph.model_utils import compress_model_weights, get_nodes_by_type
from compression.pipeline.initializer import create_pipeline
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter, create_core
from openvino.model_zoo.model_api.models import Model

from otx.algorithms.detection.adapters.openvino import model_wrappers
from otx.algorithms.detection.configs.base import DetectionConfig
from otx.api.configuration.helper.utils import config_to_bytes
from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import (
    InferenceParameters,
    default_progress_callback,
)
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.model_template import TaskType
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.serialization.label_mapper import LabelSchemaMapper, label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.exportable_code import demo
from otx.api.usecases.exportable_code.inference import BaseInferencer
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    DetectionToAnnotationConverter,
    IPredictionToAnnotationConverter,
    MaskToAnnotationConverter,
    RotatedRectToAnnotationConverter,
)
from otx.api.usecases.tasks.interfaces.deployment_interface import IDeploymentTask
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.utils import Tiler
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.dataset_utils import add_saliency_maps_to_dataset_item
from otx.api.utils.detection_utils import detection2array
from otx.mpa.utils.logger import get_logger

logger = get_logger()


# pylint: disable=too-many-locals
class BaseInferencerWithConverter(BaseInferencer):
    """BaseInferencerWithConverter class in OpenVINO task."""

    @check_input_parameters_type()
    def __init__(
        self,
        configuration: dict,
        model: Model,
        converter: IPredictionToAnnotationConverter,
    ) -> None:
        self.configuration = configuration
        self.model = model
        self.converter = converter

    @check_input_parameters_type()
    def pre_process(self, image: np.ndarray) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
        """Pre-process function of OpenVINO Detection Inferencer."""
        return self.model.preprocess(image)

    @check_input_parameters_type()
    def post_process(self, prediction: Dict[str, np.ndarray], metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Post-process function of OpenVINO Detection Inferencer."""
        detections = self.model.postprocess(prediction, metadata)

        return self.converter.convert_to_annotation(detections, metadata)

    @check_input_parameters_type()
    def predict(self, image: np.ndarray):
        """Predict function of OpenVINO Detection Inferencer."""
        image, metadata = self.pre_process(image)
        raw_predictions = self.forward(image)
        predictions = self.post_process(raw_predictions, metadata)
        if "feature_vector" not in raw_predictions or "saliency_map" not in raw_predictions:
            warnings.warn(
                "Could not find Feature Vector and Saliency Map in OpenVINO output. "
                "Please rerun OpenVINO export or retrain the model."
            )
            features = (None, None)
        else:
            features = (
                raw_predictions["feature_vector"].reshape(-1),
                raw_predictions["saliency_map"][0],
            )
        return predictions, features

    @check_input_parameters_type()
    def forward(self, image: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """Forward function of OpenVINO Detection Inferencer."""
        return self.model.infer_sync(image)

    # TODO [Eugene]: implement unittest for tiling predict
    @check_input_parameters_type()
    def predict_tile(
        self, image: np.ndarray, tile_size: int, overlap: float, max_number: int
    ) -> Tuple[AnnotationSceneEntity, Tuple[np.ndarray, np.ndarray]]:
        """Run prediction by tiling image to small patches.

        Args:
            image (np.ndarray): input image
            tile_size (int): tile crop size
            overlap (float): overlap ratio between tiles
            max_number (int): max number of predicted objects allowed

        Returns:
            detections: AnnotationSceneEntity
            features: list including saliency map and feature vector
        """
        segm = isinstance(self.converter, (MaskToAnnotationConverter, RotatedRectToAnnotationConverter))
        tiler = Tiler(tile_size=tile_size, overlap=overlap, max_number=max_number, model=self.model, segm=segm)
        detections, features = tiler.predict(image)
        detections = self.converter.convert_to_annotation(detections, metadata={"original_shape": image.shape})
        return detections, features


class OpenVINODetectionInferencer(BaseInferencerWithConverter):
    """Inferencer implementation for OTXDetection using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(
        self,
        hparams: DetectionConfig,
        label_schema: LabelSchemaEntity,
        model_file: Union[str, bytes],
        weight_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):
        """Initialize for OpenVINODetectionInferencer.

        :param hparams: Hyper parameters that the model should use.
        :param label_schema: LabelSchemaEntity that was used during model training.
        :param model_file: Path OpenVINO IR model definition file.
        :param weight_file: Path OpenVINO IR model weights file.
        :param device: Device to run inference on, such as CPU, GPU or MYRIAD. Defaults to "CPU".
        :param num_requests: Maximum number of requests that the inferencer can make. Defaults to 1.
        """

        model_adapter = OpenvinoAdapter(
            create_core(),
            model_file,
            weight_file,
            device=device,
            max_num_requests=num_requests,
        )
        configuration = {
            **attr.asdict(
                hparams.postprocessing,
                filter=lambda attr, value: attr.name not in ["header", "description", "type", "visible_in_ui"],
            )
        }
        model = Model.create_model("OTX_SSD", model_adapter, configuration, preload=True)
        converter = DetectionToAnnotationConverter(label_schema)

        super().__init__(configuration, model, converter)

    @check_input_parameters_type()
    def post_process(self, prediction: Dict[str, np.ndarray], metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Detection specific post-process."""
        detections = self.model.postprocess(prediction, metadata)
        detections = detection2array(detections)
        return self.converter.convert_to_annotation(detections, metadata)


class OpenVINOMaskInferencer(BaseInferencerWithConverter):
    """Mask Inferencer implementation for OTXDetection using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(
        self,
        hparams: DetectionConfig,
        label_schema: LabelSchemaEntity,
        model_file: Union[str, bytes],
        weight_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):
        model_adapter = OpenvinoAdapter(
            create_core(),
            model_file,
            weight_file,
            device=device,
            max_num_requests=num_requests,
        )

        configuration = {
            **attr.asdict(
                hparams.postprocessing,
                filter=lambda attr, value: attr.name not in ["header", "description", "type", "visible_in_ui"],
            )
        }

        model = Model.create_model("OTX_MaskRCNN", model_adapter, configuration, preload=True)

        converter = MaskToAnnotationConverter(label_schema)

        super().__init__(configuration, model, converter)


class OpenVINORotatedRectInferencer(BaseInferencerWithConverter):
    """Rotated Rect Inferencer implementation for OTXDetection using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(
        self,
        hparams: DetectionConfig,
        label_schema: LabelSchemaEntity,
        model_file: Union[str, bytes],
        weight_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):
        model_adapter = OpenvinoAdapter(
            create_core(),
            model_file,
            weight_file,
            device=device,
            max_num_requests=num_requests,
        )

        configuration = {
            **attr.asdict(
                hparams.postprocessing,
                filter=lambda attr, value: attr.name not in ["header", "description", "type", "visible_in_ui"],
            )
        }

        model = Model.create_model("OTX_MaskRCNN", model_adapter, configuration, preload=True)

        converter = RotatedRectToAnnotationConverter(label_schema)

        super().__init__(configuration, model, converter)


class OTXOpenVinoDataLoader(DataLoader):
    """Data loader for OTXDetection using OpenVINO backend."""

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def __init__(self, dataset: DatasetEntity, inferencer: BaseInferencer):
        self.dataset = dataset
        self.inferencer = inferencer

    @check_input_parameters_type()
    def __getitem__(self, index: int):
        """Return dataset item from index."""
        image = self.dataset[index].numpy
        annotation = self.dataset[index].annotation_scene
        inputs, metadata = self.inferencer.pre_process(image)

        return (index, annotation), inputs, metadata

    def __len__(self):
        """Length of OTXOpenVinoDataLoader."""
        return len(self.dataset)


class OpenVINODetectionTask(IDeploymentTask, IInferenceTask, IEvaluationTask, IOptimizationTask):
    """Task implementation for OTXDetection using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment):
        logger.info("Loading OpenVINO OTXDetectionTask")
        self.task_environment = task_environment
        self.model = self.task_environment.model
        self.task_type = self.task_environment.model_template.task_type
        self.confidence_threshold: float = 0.0
        self.config = self.load_config()
        self.inferencer = self.load_inferencer()
        logger.info("OpenVINO task initialization completed")

    @property
    def hparams(self):
        """Hparams of OpenVINO Detection Task."""
        return self.task_environment.get_hyper_parameters(DetectionConfig)

    def load_config(self) -> Dict:
        """Load configurable parameters from model adapter.

        Returns:
            Dict: config dictionary
        """
        try:
            if self.model is not None and self.model.get_data("config.json"):
                return json.loads(self.model.get_data("config.json"))
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"Failed to load config.json: {e}")
        return {}

    def load_inferencer(
        self,
    ) -> Union[OpenVINODetectionInferencer, OpenVINOMaskInferencer]:
        """load_inferencer function of OpenVINO Detection Task."""
        if self.model is None:
            raise RuntimeError("load_inferencer failed, model is None")
        _hparams = copy.deepcopy(self.hparams)
        self.confidence_threshold = float(
            np.frombuffer(self.model.get_data("confidence_threshold"), dtype=np.float32)[0]
        )
        _hparams.postprocessing.confidence_threshold = self.confidence_threshold
        args = [
            _hparams,
            self.task_environment.label_schema,
            self.model.get_data("openvino.xml"),
            self.model.get_data("openvino.bin"),
        ]
        if self.task_type == TaskType.DETECTION:
            return OpenVINODetectionInferencer(*args)
        if self.task_type == TaskType.INSTANCE_SEGMENTATION:
            return OpenVINOMaskInferencer(*args)
        if self.task_type == TaskType.ROTATED_DETECTION:
            return OpenVINORotatedRectInferencer(*args)
        raise RuntimeError(f"Unknown OpenVINO Inferencer TaskType: {self.task_type}")

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Infer function of OpenVINODetectionTask."""
        logger.info("Start OpenVINO inference")

        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress
            add_saliency_map = not inference_parameters.is_evaluation
            process_saliency_maps = inference_parameters.process_saliency_maps
            explain_predicted_classes = inference_parameters.explain_predicted_classes
        else:
            update_progress_callback = default_progress_callback
            add_saliency_map = True
            process_saliency_maps = False
            explain_predicted_classes = True

        tile_enabled = bool(self.config and self.config["tiling_parameters"]["enable_tiling"]["value"])

        if tile_enabled:
            tile_size = self.config["tiling_parameters"]["tile_size"]["value"]
            tile_overlap = self.config["tiling_parameters"]["tile_overlap"]["value"]
            max_number = self.config["tiling_parameters"]["tile_max_number"]["value"]
            logger.info("Run inference with tiling")

        dataset_size = len(dataset)
        for i, dataset_item in enumerate(dataset, 1):
            if tile_enabled:
                predicted_scene, features = self.inferencer.predict_tile(
                    dataset_item.numpy,
                    tile_size=tile_size,
                    overlap=tile_overlap,
                    max_number=max_number,
                )
            else:
                predicted_scene, features = self.inferencer.predict(dataset_item.numpy)

            dataset_item.append_annotations(predicted_scene.annotations)
            feature_vector, saliency_map = features
            if feature_vector is not None:
                representation_vector = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(representation_vector, model=self.model)

            if add_saliency_map and saliency_map is not None:
                labels = self.task_environment.get_labels().copy()
                if saliency_map.shape[0] == len(labels) + 1:
                    # Include the background as the last category
                    labels.append(LabelEntity("background", Domain.DETECTION))

                predicted_scored_labels = []
                for bbox in predicted_scene.annotations:
                    predicted_scored_labels += bbox.get_labels()

                add_saliency_maps_to_dataset_item(
                    dataset_item=dataset_item,
                    saliency_map=saliency_map,
                    model=self.model,
                    labels=labels,
                    predicted_scored_labels=predicted_scored_labels,
                    explain_predicted_classes=explain_predicted_classes,
                    process_saliency_maps=process_saliency_maps,
                )
            update_progress_callback(int(i / dataset_size * 100), None)
        logger.info("OpenVINO inference completed")
        return dataset

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Explain function of OpenVINODetectionTask."""
        logger.info("Start OpenVINO explain")

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if explain_parameters is not None:
            update_progress_callback = explain_parameters.update_progress  # type: ignore
            process_saliency_maps = explain_parameters.process_saliency_maps
            explain_predicted_classes = explain_parameters.explain_predicted_classes

        dataset_size = len(dataset)
        for i, dataset_item in enumerate(dataset, 1):
            predicted_scene, features = self.inferencer.predict(dataset_item.numpy)
            dataset_item.append_annotations(predicted_scene.annotations)
            update_progress_callback(int(i / dataset_size * 100), None)
            _, saliency_map = features

            labels = self.task_environment.get_labels().copy()
            if saliency_map.shape[0] == len(labels) + 1:
                # Include the background as the last category
                labels.append(LabelEntity("background", Domain.DETECTION))

            predicted_scored_labels = []
            for bbox in predicted_scene.annotations:
                predicted_scored_labels += bbox.get_labels()

            add_saliency_maps_to_dataset_item(
                dataset_item=dataset_item,
                saliency_map=np.copy(saliency_map),
                model=self.model,
                labels=labels,
                predicted_scored_labels=predicted_scored_labels,
                explain_predicted_classes=explain_predicted_classes,
                process_saliency_maps=process_saliency_maps,
            )
        logger.info("OpenVINO explain completed")
        return dataset

    @check_input_parameters_type()
    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OpenVINODetectionTask."""
        logger.info("Start OpenVINO metric evaluation")
        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric, but parameter is ignored. Use F-measure instead."
            )
        output_resultset.performance = MetricsHelper.compute_f_measure(output_resultset).get_performance()
        logger.info("OpenVINO metric evaluation completed")

    @check_input_parameters_type()
    def deploy(self, output_model: ModelEntity) -> None:
        """Deploy function of OpenVINODetectionTask."""
        logger.info("Deploying the model")

        work_dir = os.path.dirname(demo.__file__)
        parameters = {}
        parameters["type_of_model"] = self.inferencer.model.__model__
        parameters["converter_type"] = str(self.task_type)
        parameters["model_parameters"] = self.inferencer.configuration
        parameters["model_parameters"]["labels"] = LabelSchemaMapper.forward(self.task_environment.label_schema)
        parameters["tiling_parameters"] = self.config["tiling_parameters"]

        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, "w") as arch:
            # model files
            if self.model is None:
                raise ValueError("Deploy failed, model is None")
            arch.writestr(os.path.join("model", "model.xml"), self.model.get_data("openvino.xml"))
            arch.writestr(os.path.join("model", "model.bin"), self.model.get_data("openvino.bin"))
            arch.writestr(
                os.path.join("model", "config.json"),
                json.dumps(parameters, ensure_ascii=False, indent=4),
            )
            # model_wrappers files
            for root, _, files in os.walk(os.path.dirname(model_wrappers.__file__)):
                for file in files:
                    file_path = os.path.join(root, file)
                    arch.write(
                        file_path,
                        os.path.join(
                            "python",
                            "model_wrappers",
                            file_path.split("model_wrappers/")[1],
                        ),
                    )
            # python files
            arch.write(
                os.path.join(work_dir, "requirements.txt"),
                os.path.join("python", "requirements.txt"),
            )
            arch.write(os.path.join(work_dir, "LICENSE"), os.path.join("python", "LICENSE"))
            arch.write(os.path.join(work_dir, "README.md"), os.path.join("python", "README.md"))
            arch.write(os.path.join(work_dir, "demo.py"), os.path.join("python", "demo.py"))
        output_model.exportable_code = zip_buffer.getvalue()
        logger.info("Deploying completed")

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        """Optimize function of OpenVINODetectionTask."""
        logger.info("Start POT optimization")

        if optimization_type is not OptimizationType.POT:
            raise ValueError("POT is the only supported optimization type for OpenVino models")
        if self.model is None:
            raise RuntimeError("Optimize failed, model is None")

        dataset = dataset.get_subset(Subset.TRAINING)
        data_loader = OTXOpenVinoDataLoader(dataset, self.inferencer)

        with tempfile.TemporaryDirectory() as tempdir:
            xml_path = os.path.join(tempdir, "model.xml")
            bin_path = os.path.join(tempdir, "model.bin")
            with open(xml_path, "wb") as f:
                f.write(self.model.get_data("openvino.xml"))
            with open(bin_path, "wb") as f:
                f.write(self.model.get_data("openvino.bin"))

            model_config = ADDict({"model_name": "openvino_model", "model": xml_path, "weights": bin_path})

            model = load_model(model_config)

            if get_nodes_by_type(model, ["FakeQuantize"]):
                raise RuntimeError("Model is already optimized by POT")

        if optimization_parameters:
            optimization_parameters.update_progress(10, None)

        engine_config = ADDict(
            {
                "device": "CPU",
                "stat_requests_number": min(
                    self.hparams.pot_parameters.stat_requests_number, multiprocessing.cpu_count()
                ),
            }
        )

        stat_subset_size = self.hparams.pot_parameters.stat_subset_size
        preset = self.hparams.pot_parameters.preset.name.lower()

        algorithms = [
            {
                "name": "DefaultQuantization",
                "params": {
                    "target_device": "ANY",
                    "preset": preset,
                    "stat_subset_size": min(stat_subset_size, len(data_loader)),
                    "shuffle_data": True,
                },
            }
        ]

        engine = IEEngine(config=engine_config, data_loader=data_loader, metric=None)

        pipeline = create_pipeline(algorithms, engine)

        compressed_model = pipeline.run(model)

        compress_model_weights(compressed_model)

        if optimization_parameters:
            optimization_parameters.update_progress(90, None)

        with tempfile.TemporaryDirectory() as tempdir:
            save_model(compressed_model, tempdir, model_name="model")
            with open(os.path.join(tempdir, "model.xml"), "rb") as f:
                output_model.set_data("openvino.xml", f.read())
            with open(os.path.join(tempdir, "model.bin"), "rb") as f:
                output_model.set_data("openvino.bin", f.read())
            output_model.set_data(
                "confidence_threshold",
                np.array([self.confidence_threshold], dtype=np.float32).tobytes(),
            )

        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self.task_environment.label_schema),
        )
        output_model.set_data("config.json", config_to_bytes(self.hparams))

        # set model attributes for quantized model
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.POT
        output_model.optimization_methods = [OptimizationMethod.QUANTIZATION]
        output_model.precision = [ModelPrecision.INT8]

        self.model = output_model
        self.inferencer = self.load_inferencer()
        logger.info("POT optimization completed")

        if optimization_parameters:
            optimization_parameters.update_progress(100, None)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa
from . import (
    adaptive_training_hooks,
    checkpoint_hook,
    composed_dataloaders_hook,
    early_stopping_hook,
    fp16_sam_optimizer_hook,
    ib_loss_hook,
    logger_replace_hook,
    model_ema_hook,
    model_ema_v2_hook,
    no_bias_decay_hook,
    recording_forward_hooks,
    sam_optimizer_hook,
    save_initial_weight_hook,
    semisl_cls_hook,
    task_adapt_hook,
    unbiased_teacher_hook,
    workflow_hooks,
)



"""Callback module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


class Callback:
    """Abstract base class used to build new callbacks.

    Properties
        params: dict. Training parameters
            (eg. verbosity, batch size, number of epochs...).
        model: instance of `keras.models.Model`.
            Reference of the model being trained.

    The `logs` dictionary that callback methods
    take as argument will contain keys for quantities relevant to
    the current batch or epoch.

    Currently, the `.fit()` method of the `Sequential` model class
    will include the following quantities in the `logs` that
    it passes to its callbacks:

        on_epoch_end: logs include `acc` and `loss`, and
            optionally include `val_loss`
            (if validation is enabled in `fit`), and `val_acc`
            (if validation and accuracy monitoring are enabled).
        on_batch_begin: logs include `size`,
            the number of samples in the current batch.
        on_batch_end: logs include `loss`, and optionally `acc`
            (if accuracy monitoring is enabled).
    """

    def set_params(self, params):
        """Sets callback parameters."""
        # pylint: disable=W0201
        self.params = params

    def set_model(self, model):
        """Sets callback model."""
        # pylint: disable=W0201
        self.model = model

    def on_epoch_begin(self, epoch, logs=None):
        """It is called on epoch begin event."""

    def on_epoch_end(self, epoch, logs=None):
        """It is called on epoch end event."""

    def on_batch_begin(self, batch, logs=None):
        """It is called on batch begin event."""

    def on_batch_end(self, batch, logs=None):
        """It is called on batch end event."""

    def on_train_begin(self, logs=None):
        """It is called on train begin event."""

    def on_train_end(self, logs=None):
        """It is called on train end event."""

    def on_train_batch_begin(self, batch, logs):
        """It is called on train batch begin event."""

    def on_train_batch_end(self, batch, logs):
        """It is called on train batch end event."""

    def on_test_begin(self, logs):
        """It is called on test begin event."""

    def on_test_end(self, logs):
        """It is called on test end event."""

    def on_test_batch_begin(self, batch, logs):
        """It is called on test batch begin event."""

    def on_test_batch_end(self, batch, logs):
        """It is called on test batch end event."""



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.utils import get_logger

from otx.algorithms.common.adapters.mmcv.nncf.hooks import CompressionHook
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import (
    create_nncf_model,
)


class TestCompressionHook:
    @e2e_pytest_unit
    def test_after_train_iter(self):
        class SimpleRunner:
            def __init__(self):
                self.logger = get_logger("mmcv")
                self.rank = 0

        runner = SimpleRunner()
        ctrl, _ = create_nncf_model()
        compression_hook = CompressionHook(ctrl)
        compression_hook.after_train_iter(runner)

    @e2e_pytest_unit
    def test_after_train_epoch(self):
        class SimpleRunner:
            def __init__(self):
                self.logger = get_logger("mmcv")
                self.rank = 0

        runner = SimpleRunner()
        ctrl, _ = create_nncf_model()
        compression_hook = CompressionHook(ctrl)
        compression_hook.after_train_epoch(runner)

    @e2e_pytest_unit
    def test_before_run(self):
        class SimpleRunner:
            def __init__(self):
                self.logger = get_logger("mmcv")
                self.rank = 0

        runner = SimpleRunner()
        ctrl, _ = create_nncf_model()
        compression_hook = CompressionHook(ctrl)
        compression_hook.before_run(runner)



"""Detection Utils."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from typing import Dict, List

import cv2
import numpy as np

from otx.api.entities.annotation import Annotation
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle


def create_detection_annotation_from_anomaly_heatmap(
    hard_prediction: np.ndarray,
    soft_prediction: np.ndarray,
    label_map: Dict[int, LabelEntity],
) -> List[Annotation]:
    """Create box annotation from the soft predictions.

    Args:
        hard_prediction: hard prediction containing the final label
            index per pixel.
        soft_prediction: soft prediction with shape
        label_map: dictionary mapping labels to an index. It is assumed
            that the first item in the dictionary corresponds to the
            background label and will therefore be ignored.

    Returns:
        List of annotations.
    """
    # pylint: disable=too-many-locals
    if hard_prediction.ndim == 3 and hard_prediction.shape[0] == 1:
        hard_prediction = hard_prediction.squeeze().astype(np.uint8)

    if soft_prediction.ndim == 3 and soft_prediction.shape[0] == 1:
        soft_prediction = soft_prediction.squeeze()

    image_h, image_w = hard_prediction.shape[:2]

    annotations: List[Annotation] = []
    for label_index, label in label_map.items():
        # Skip the normal label.
        if label_index == 0:
            continue

        # cv2.connectedComponentsWithStats returns num_labels, labels, coordinates
        # and centroids. This script only needs the coordinates.
        _, connected_components, coordinates, _ = cv2.connectedComponentsWithStats(hard_prediction)

        for i, coordinate in enumerate(coordinates):
            # First row of the coordinates is always backround,
            # so should be ignored.
            if i == 0:
                continue

            # Last column of the coordinates is the area of the connected component.
            # It could therefore be ignored.
            comp_x, comp_y, comp_w, comp_h, _ = coordinate

            # Compute the probability of each connected-component
            component_hard_prediction = (connected_components == i).astype(np.uint8)
            component_soft_prediction = cv2.bitwise_and(
                soft_prediction, soft_prediction, mask=component_hard_prediction
            )

            # NOTE: Find the best approach to calculate the probability
            probability = component_soft_prediction.reshape(-1).max()

            # NOTE: NMS could be needed here.

            # Create the annotation based on the box shape and the probability.
            shape = Rectangle(
                x1=comp_x / image_w,
                y1=comp_y / image_h,
                x2=(comp_x + comp_w) / image_w,
                y2=(comp_y + comp_h) / image_h,
            )
            annotations.append(Annotation(shape=shape, labels=[ScoredLabel(label, float(probability))]))
    return annotations



"""Base MMDataset for Action Recognition Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from copy import copy
from typing import Any, Dict, List, Sequence

import numpy as np
from mmaction.datasets.builder import DATASETS
from mmaction.datasets.pipelines import Compose
from mmaction.datasets.rawframe_dataset import RawframeDataset

from otx.algorithms.action.adapters.mmaction.data.pipelines import RawFrameDecode
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)


# pylint: disable=too-many-instance-attributes
@DATASETS.register_module()
class OTXActionClsDataset(RawframeDataset):
    """Wrapper that allows using a OTX dataset to train mmaction models.

    This wrapper is not based on the filesystem,
    but instead loads the items here directly from the OTX DatasetEntity object.
    """

    class _DataInfoProxy:
        def __init__(self, otx_dataset: DatasetEntity, labels: List[LabelEntity], modality: str):
            self.otx_dataset = otx_dataset
            self.labels = labels
            self.label_idx = {label.id: i for i, label in enumerate(labels)}
            self.modality = modality
            self.video_info: Dict[str, Any] = {}
            self._update_meta_data()

        def __len__(self) -> int:
            return len(self.video_info)

        def _update_meta_data(self):
            """Update video metadata of each item in self.otx_dataset.

            This function assumes that DatasetItemEntities in DatasetEntity are sorted by video id and frame idx
            During iterating DatasetsetEntitiy, this fucnction generates video_info(dictionary)
            video_info records metadata for each video, and it contains
                - total_frame: Total frame number of the video, this value will be used to sample frames for training
                - start_index: Offset for the video, this value will be added to sampled frame indices for the video
                - label: Action category of the video
                - modality = Modality of data, 'RGB' or 'Flow(Optical Flow)'
            """
            video_info = {}
            start_index = 0
            for idx, item in enumerate(self.otx_dataset):
                metadata = item.get_metadata()[0].data
                if metadata.video_id in video_info:
                    video_info[metadata.video_id]["total_frames"] += 1
                    video_info[metadata.video_id]["start_index"] = start_index
                else:
                    if len(item.get_annotations()) > 0:
                        label = int(item.get_roi_labels(self.labels)[0].id)
                    else:
                        label = None
                    ignored_labels = np.array([self.label_idx[label.id] for label in item.ignored_labels])
                    video_info[metadata.video_id] = {
                        "total_frames": 1,
                        "start_index": idx,
                        "label": label,
                        "ignored_labels": ignored_labels,
                        "modality": self.modality,
                    }
                    start_index = idx

            self.video_info.update(video_info)

        def __getitem__(self, index: int) -> Dict[str, Any]:
            """Prepare training data item.

            Action classification needs video for training, therefore this function generate item from video_info
            """

            item = self.video_info[list(self.video_info.keys())[index]]
            return item

    @check_input_parameters_type({"otx_dataset": DatasetParamTypeCheck})
    # pylint: disable=too-many-arguments, invalid-name, super-init-not-called
    # TODO Check need for additional params such as multi_class, with_offset
    def __init__(
        self,
        otx_dataset: DatasetEntity,
        labels: List[LabelEntity],
        pipeline: Sequence[dict],
        test_mode: bool = False,
        modality: str = "RGB",  # [RGB, FLOW(Optical flow)]
    ):
        self.otx_dataset = otx_dataset
        self.labels = labels
        self.test_mode = test_mode
        self.modality = modality

        self.video_infos = OTXActionClsDataset._DataInfoProxy(otx_dataset, labels, modality)

        self.pipeline = Compose(pipeline)
        for transform in self.pipeline.transforms:
            if isinstance(transform, RawFrameDecode):
                transform.otx_dataset = self.otx_dataset

    def __len__(self) -> int:
        """Return length of dataset."""
        return len(self.video_infos)

    @check_input_parameters_type()
    def prepare_train_frames(self, idx: int) -> Dict[str, Any]:
        """Get training data and annotations after pipeline.

        Args:
            idx (int): Index of data
        """
        item = copy(self.video_infos[idx])  # Copying dict(), not contents
        return self.pipeline(item)

    @check_input_parameters_type()
    def prepare_test_frames(self, idx: int) -> Dict[str, Any]:
        """Get testing data after pipeline.

        Args:
            idx (int): Index of data
        """
        item = copy(self.video_infos[idx])  # Copying dict(), not contents
        return self.pipeline(item)



"""MobileNet-V3-large-075 for multi-class config."""

# pylint: disable=invalid-name

_base_ = "../base/models/mobilenet_v3.py"

model = dict(
    type="SAMImageClassifier",
    task="classification",
    backbone=dict(
        mode="large",
        width_mult=0.75,
    ),
    head=dict(
        in_channels=720,
        hid_channels=1280,
    ),
)



"""This module implements the ResultSet entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import abc
import datetime
from enum import Enum
from typing import Optional

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.metrics import NullPerformance, Performance
from otx.api.entities.model import ModelEntity
from otx.api.utils.time_utils import now


class ResultsetPurpose(Enum):
    """This defines the purpose of the resultset.

    EVALUATION denotes resultsets generated at Evaluation stage on validation subset.

    TEST denotes resultsets generated at Evaluation stage on test subset.

    PREEVALUATION denotes resultsets generated at Preevaluation stage (e.g., train from
    scratch) onn validation subset.
    """

    EVALUATION = 0
    TEST = 1
    PREEVALUATION = 2

    def __repr__(self):
        """Returns ResultsetPurpose as a string."""
        return str(self.name)

    def __str__(self):
        """Returns a user friendly representation of the ResultSetPurpose.

        This that can be used for instance in a progress reporting message.
        """
        user_friendly_names = {0: "Validation", 1: "Test", 2: "Pre-validation"}
        return user_friendly_names[self.value]


class ResultSetEntity(metaclass=abc.ABCMeta):
    """ResultsetEntity.

    It aggregates:
        - the dataset containing ground truth (based on user annotations)
        - the dataset containing predictions for the above ground truth dataset

    In addition, it links to the model which computed the predictions, as well as the performance of this model on the
    ground truth dataset.

    Args:
        model: the model using which the prediction_dataset has been
            generated
        ground_truth_dataset: the dataset containing ground truth
            annotation
        prediction_dataset: the dataset containing prediction
        purpose: see :class:`ResultsetPurpose`
        performance: the performance of the model on the ground truth
            dataset
        creation_date: the date time which the resultset is created. Set
            to None to set this to
        id: the id of the resultset. Set to ID() so that a new unique ID
            will be assigned upon saving. If the argument is None, it
            will be set to ID()
    datetime.now(datetime.timezone.utc)
    """

    # pylint: disable=redefined-builtin, too-many-arguments; Requires refactor
    def __init__(
        self,
        model: ModelEntity,
        ground_truth_dataset: DatasetEntity,
        prediction_dataset: DatasetEntity,
        purpose: ResultsetPurpose = ResultsetPurpose.EVALUATION,
        performance: Optional[Performance] = None,
        creation_date: Optional[datetime.datetime] = None,
        id: Optional[ID] = None,
    ):
        id = ID() if id is None else id
        performance = NullPerformance() if performance is None else performance
        creation_date = now() if creation_date is None else creation_date
        self.__id_ = id
        self.__model = model
        self.__prediction_dataset = prediction_dataset
        self.__ground_truth_dataset = ground_truth_dataset
        self.__performance = performance
        self.__purpose = purpose
        self.__creation_date = creation_date

    @property
    def id_(self) -> ID:
        """Returns the id of the ResultSet."""
        return self.__id_

    @id_.setter
    def id_(self, value: ID) -> None:
        self.__id_ = value

    @property
    def id(self) -> ID:
        """DEPRECATED."""
        return self.__id_

    @id.setter
    def id(self, value: ID):
        """DEPRECATED."""
        self.__id_ = value

    @property
    def model(self) -> ModelEntity:
        """Returns the model that is used for the ResultSet."""
        return self.__model

    @model.setter
    def model(self, value: ModelEntity) -> None:
        self.__model = value

    @property
    def prediction_dataset(self) -> DatasetEntity:
        """Returns the prediction dataset that is used in the ResultSet."""
        return self.__prediction_dataset

    @prediction_dataset.setter
    def prediction_dataset(self, value: DatasetEntity) -> None:
        self.__prediction_dataset = value

    @property
    def ground_truth_dataset(self) -> DatasetEntity:
        """Returns the ground truth dataset that is used in the ResultSet."""
        return self.__ground_truth_dataset

    @ground_truth_dataset.setter
    def ground_truth_dataset(self, value: DatasetEntity) -> None:
        self.__ground_truth_dataset = value

    @property
    def performance(self) -> Performance:
        """Returns the performance of the model on the ground truth dataset."""
        return self.__performance

    @performance.setter
    def performance(self, value: Performance) -> None:
        self.__performance = value

    @property
    def purpose(self) -> ResultsetPurpose:
        """Returns the purpose of the ResultSet, for example ResultSetPurpose.EVALUATION."""
        return self.__purpose

    @purpose.setter
    def purpose(self, value: ResultsetPurpose) -> None:
        self.__purpose = value

    @property
    def creation_date(self) -> datetime.datetime:
        """Returns the creation date of the ResultSet."""
        return self.__creation_date

    @creation_date.setter
    def creation_date(self, value: datetime.datetime) -> None:
        self.__creation_date = value

    def has_score_metric(self) -> bool:
        """Returns True if the resultset contains non-null performance and score value."""
        return not isinstance(self.performance, NullPerformance)

    def __repr__(self):
        """String representation of the resultset."""
        return (
            f"{type(self).__name__}("
            f"model={self.model}, "
            f"ground_truth_dataset={self.ground_truth_dataset}, "
            f"prediction_dataset={self.prediction_dataset}, "
            f"purpose={self.purpose}, "
            f"performance={self.performance}, "
            f"creation_date={self.creation_date}, "
            f"id={self.id_})"
        )



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import random

from mmcls.datasets.builder import PIPELINES
from PIL import Image
from torchvision.transforms import functional as F


@PIPELINES.register_module()
class PILToTensor(object):
    def __call__(self, results):
        for key in results.get("img_fields", ["img"]):
            img = results[key]
            if not Image.isImageType(img):
                img = Image.fromarray(img)
            img = F.to_tensor(img)
            results["PILToTensor"] = True
            results[key] = img
        return results


@PIPELINES.register_module()
class TensorNormalize(object):
    def __init__(self, mean, std, inplace=False):
        self.mean = mean
        self.std = std
        self.inplace = inplace

    def __call__(self, results):
        for key in results.get("img_fields", ["img"]):
            img = results[key]
            img = F.normalize(img, self.mean, self.std, self.inplace)
            results["TensorNormalize"] = True
            results[key] = img
        return results


# TODO [Jihwan]: Can be removed by mmcls.dataset.pipelines.auto_augment L398, Roate class
@PIPELINES.register_module()
class RandomRotate(object):
    """Random rotate
    From torchreid.data.transforms
    """

    def __init__(self, p=0.5, angle=(-5, 5), values=None, **kwargs):
        self.p = p
        self.angle = angle
        self.discrete = values is not None and len([v for v in values if v != 0]) > 0
        self.values = values

    def __call__(self, results, *args, **kwargs):
        if random.uniform(0, 1) > self.p:
            return results
        for key in results.get("img_fields", ["img"]):
            img = results[key]
            if self.discrete:
                rnd_angle = float(self.values[random.randint(0, len(self.values) - 1)])
            else:
                rnd_angle = random.randint(self.angle[0], self.angle[1])
            if not Image.isImageType(img):
                img = Image.fromarray(img)

            img = F.rotate(img, rnd_angle, expand=False, center=None)
            results["RandomRotate"] = True
            results[key] = img

        return results



"""Synchronous Executor based on ModelAPI."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Union

from otx.api.usecases.exportable_code.demo.demo_package.model_container import (
    ModelContainer,
)
from otx.api.usecases.exportable_code.demo.demo_package.utils import (
    create_output_converter,
)
from otx.api.usecases.exportable_code.streamer import get_streamer
from otx.api.usecases.exportable_code.visualizers import Visualizer


class SyncExecutor:
    """Synchronous executor for model inference.

    Args:
        model (ModelContainer): model for inference
        visualizer (Visualizer): visualizer of inference results. Defaults to None.
    """

    def __init__(self, model: ModelContainer, visualizer: Visualizer) -> None:
        self.model = model
        self.visualizer = visualizer
        self.converter = create_output_converter(model.task_type, model.labels)

    def run(self, input_stream: Union[int, str], loop: bool = False) -> None:
        """Run demo using input stream (image, video stream, camera)."""
        streamer = get_streamer(input_stream, loop)

        for frame in streamer:
            # getting result include preprocessing, infer, postprocessing for sync infer
            predictions, frame_meta = self.model(frame)
            annotation_scene = self.converter.convert_to_annotation(predictions, frame_meta)
            output = self.visualizer.draw(frame, annotation_scene, frame_meta)
            self.visualizer.show(output)
            if self.visualizer.is_quit():
                break



"""Data Pipeline of Resnet model for Rotated-Detection Task."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name

__img_size = (1344, 800)

# TODO: A comparison experiment is needed to determine which value is appropriate for to_rgb.
__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True, with_mask=True, poly2mask=False),
    dict(type="Resize", img_scale=__img_size, keep_ratio=False),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels", "gt_masks"]),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_size,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]

__dataset_type = "CocoDataset"

data = dict(
    samples_per_gpu=4,
    workers_per_gpu=2,
    train=dict(
        type=__dataset_type,
        ann_file="data/coco/annotations/instances_train2017.json",
        img_prefix="data/coco/train2017",
        pipeline=train_pipeline,
    ),
    val=dict(
        type=__dataset_type,
        test_mode=True,
        ann_file="data/coco/annotations/instances_val2017.json",
        img_prefix="data/coco/val2017",
        pipeline=test_pipeline,
    ),
    test=dict(
        type=__dataset_type,
        test_mode=True,
        ann_file="data/coco/annotations/instances_val2017.json",
        img_prefix="data/coco/val2017",
        pipeline=test_pipeline,
    ),
)



"""This module define a module to adapt model weights from a data source."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from typing import Union


class IDataSource:
    """Class that holds a combination of both a repo and a URL which can be used to fetch data."""

    @property
    @abc.abstractmethod
    def data(self):
        """Returns the data of the source."""
        raise NotImplementedError


class ModelAdapter(metaclass=abc.ABCMeta):
    """The ModelAdapter is an adapter is intended to lazily fetch its binary data from a given data source."""

    def __init__(self, data_source: Union[IDataSource, bytes]):
        self.__data_source = data_source

    @property
    def data_source(self):
        """Returns the data source of the adapter."""
        return self.__data_source

    @data_source.setter
    def data_source(self, value: Union[IDataSource, bytes]):
        self.__data_source = value

    @property
    def data(self):
        """Returns the data of the Model."""
        if isinstance(self.__data_source, IDataSource):
            return self.__data_source.data
        if isinstance(self.__data_source, bytes):
            return self.__data_source
        raise ValueError("This model adapter is not properly initialized with a source of data")

    @property
    def from_file_storage(self) -> bool:
        """Returns if the ModelAdapters data comes from the file storage or not.

        This is used in the model repo to know if the data of the model should be saved or not.
        """
        if isinstance(self.data_source, bytes):
            return False
        return True


class ExportableCodeAdapter(ModelAdapter):
    """Adapter intended to lazily fetch raw exportable code data from a given data source."""



"""This module contains the interface class for tasks that can optimize their models."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from enum import Enum, auto
from typing import Optional

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.optimization_parameters import OptimizationParameters


class OptimizationType(Enum):
    """This class enumerates the OPENVINO optimization types."""

    POT = auto()
    NNCF = auto()


class IOptimizationTask(metaclass=abc.ABCMeta):
    """A base interface class for tasks which can optimize their models."""

    @abc.abstractmethod
    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters],
    ):
        """This method defines the interface for optimization.

        Args:
            optimization_type: The type of optimization
            dataset: Optional dataset which may be used as part of the
                optimization process
            output_model: Output model
            optimization_parameters: Additional optimization parameters
        """
        raise NotImplementedError



"""OTX configurable parameters and helper utilities.

This module contains base elements that make up OTX ConfigurableParameters, as well as a collection of helper
functions to interact with them.

The configuration helper module can be imported as `otx_config_helper` and implements the following:

.. automodule:: otx.api.configuration.helper
   :members:

"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import otx.api.configuration.helper as ote_config_helper  # for 'ote' backward compatibility
import otx.api.configuration.helper as cfg_helper  # pylint: disable=reimported
from otx.api.configuration.elements import metadata_keys
from otx.api.configuration.elements.configurable_enum import ConfigurableEnum
from otx.api.configuration.enums.model_lifecycle import ModelLifecycle
from otx.api.configuration.ui_rules import Action, NullUIRules, Operator, Rule, UIRules

from .configurable_parameters import ConfigurableParameters
from .default_model_parameters import DefaultModelParameters

__all__ = [
    "metadata_keys",
    "cfg_helper",
    "ote_config_helper",
    "ConfigurableEnum",
    "ModelLifecycle",
    "Action",
    "NullUIRules",
    "Operator",
    "Rule",
    "UIRules",
    "DefaultModelParameters",
    "ConfigurableParameters",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import traceback

from otx.mpa.utils.logger import get_logger

logger = get_logger()


class ExporterMixin(object):
    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):  # noqa: C901
        self._init_logger()
        logger.info("exporting the model")
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("export!")

        #  from torch.jit._trace import TracerWarning
        #  import warnings
        #  warnings.filterwarnings("ignore", category=TracerWarning)
        precision = kwargs.pop("precision", "FP32")
        if precision not in ("FP32", "FP16", "INT8"):
            raise NotImplementedError
        logger.info(f"Model will be exported with precision {precision}")
        model_name = cfg.get("model_name", "model")

        # TODO: handle complicated pipeline
        # If test dataset is a wrapper dataset
        # pipeline may not include load transformation which is assumed to be included afterwards
        # Here, we assume simple wrapper datasets where pipeline of the wrapper is just a consecutive one.
        if cfg.data.test.get("dataset", None) or cfg.data.test.get("datasets", None):
            dataset = cfg.data.test.get("dataset", cfg.data.test.get("datasets", [None])[0])
            assert dataset is not None
            pipeline = dataset.get("pipeline", [])
            pipeline += cfg.data.test.get("pipeline", [])
            cfg.data.test.pipeline = pipeline

        model_builder = kwargs.get("model_builder")
        try:
            deploy_cfg = kwargs.get("deploy_cfg", None)
            if deploy_cfg is not None:
                self.mmdeploy_export(
                    cfg.work_dir,
                    model_builder,
                    precision,
                    cfg,
                    deploy_cfg,
                    model_name,
                )
            else:
                self.naive_export(cfg.work_dir, model_builder, precision, cfg, model_name)
        except Exception as ex:
            # output_model.model_status = ModelStatus.FAILED
            # raise RuntimeError('Optimization was unsuccessful.') from ex
            return {
                "outputs": None,
                "msg": f"exception {type(ex)}: {ex}\n\n{traceback.format_exc()}",
            }

        return {
            "outputs": {
                "bin": os.path.join(cfg.work_dir, f"{model_name}.bin"),
                "xml": os.path.join(cfg.work_dir, f"{model_name}.xml"),
                "partitioned": [
                    {
                        "bin": os.path.join(cfg.work_dir, name.replace(".xml", ".bin")),
                        "xml": os.path.join(cfg.work_dir, name),
                    }
                    for name in os.listdir(cfg.work_dir)
                    if name.endswith(".xml")
                    and name != f"{model_name}.xml"
                    and name.replace(".xml", ".bin") in os.listdir(cfg.work_dir)
                ],
            },
            "msg": "",
        }

    @staticmethod
    def mmdeploy_export(
        output_dir,
        model_builder,
        precision,
        cfg,
        deploy_cfg,
        model_name="model",
    ):
        from .deploy.apis import MMdeployExporter

        if precision == "FP16":
            deploy_cfg.backend_config.mo_options.flags.append("--compress_to_fp16")
        MMdeployExporter.export2openvino(output_dir, model_builder, cfg, deploy_cfg, model_name=model_name)

    @staticmethod
    def naive_export(output_dir, model_builder, precision, cfg, model_name="model"):
        raise NotImplementedError()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field

import torch

from .builder import OPS
from .op import Attribute, Operation
from .type_conversions import ConvertV0


@dataclass
class SqueezeV0Attribute(Attribute):
    pass


@OPS.register()
class SqueezeV0(Operation[SqueezeV0Attribute]):
    TYPE = "Squeeze"
    VERSION = 0
    ATTRIBUTE_FACTORY = SqueezeV0Attribute

    def forward(self, input, dims=None):
        if dims is None:
            return torch.squeeze(input)

        if dims.dim() == 0:
            dims = torch.unsqueeze(dims, 0)

        max_dim = input.dim()
        dims = dims.detach().cpu().tolist()
        for i, dim in enumerate(dims):
            if dim < 0:
                dims[i] = max_dim + dim

        output = input
        for dim in sorted(dims, reverse=True):
            output = torch.squeeze(output, dim)

        return output


@dataclass
class UnsqueezeV0Attribute(Attribute):
    pass


@OPS.register()
class UnsqueezeV0(Operation[UnsqueezeV0Attribute]):
    TYPE = "Unsqueeze"
    VERSION = 0
    ATTRIBUTE_FACTORY = UnsqueezeV0Attribute

    def forward(self, input, dims):

        if dims.dim() == 0:
            dims = torch.unsqueeze(dims, 0)

        max_dim = input.dim()
        dims = dims.detach().cpu().tolist()
        if len(dims) > 1:
            for i, dim in enumerate(dims):
                if dim < 0:
                    dims[i] = max_dim + dim

        output = input
        for dim in sorted(dims, reverse=True):
            output = torch.unsqueeze(output, dim)

        return output


@dataclass
class ReshapeV1Attribute(Attribute):
    special_zero: bool


@OPS.register()
class ReshapeV1(Operation[ReshapeV1Attribute]):
    TYPE = "Reshape"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReshapeV1Attribute

    def forward(self, input, shape):
        target_shape = shape.detach().cpu().tolist()
        origin_shape = list(input.shape)
        for i, (origin_dim, target_dim) in enumerate(zip(origin_shape, target_shape)):
            if target_dim == 0 and self.attrs.special_zero:
                target_shape[i] = origin_dim
            elif target_dim == -1:
                break
        for i, (origin_dim, target_dim) in enumerate(zip(origin_shape[::-1], target_shape[::-1])):
            if target_dim == 0 and self.attrs.special_zero:
                target_shape[i] = origin_dim
            elif target_dim == -1:
                break
        return torch.reshape(input, target_shape)


@dataclass
class ShapeOfV0Attribute(Attribute):
    pass


@OPS.register()
class ShapeOfV0(Operation[ShapeOfV0Attribute]):
    TYPE = "ShapeOf"
    VERSION = 0
    ATTRIBUTE_FACTORY = ShapeOfV0Attribute

    def forward(self, input):
        return torch.tensor(input.shape, device=input.device)


@dataclass
class ShapeOfV3Attribute(Attribute):
    output_type: str = field(default="i64")

    def __post_init__(self):
        super().__post_init__()
        valid_output_type = ["i64", "i32"]
        if self.output_type not in valid_output_type:
            raise ValueError(f"Invalid output_type {self.output_type}. " f"It must be one of {valid_output_type}.")


@OPS.register()
class ShapeOfV3(Operation[ShapeOfV3Attribute]):
    TYPE = "ShapeOf"
    VERSION = 3
    ATTRIBUTE_FACTORY = ShapeOfV3Attribute

    def forward(self, input):
        return ConvertV0("temp", shape=self.shape, destination_type=self.attrs.output_type)(
            torch.tensor(input.shape, device=input.device)
        )



"""Unit Test for otx.algorithms.action.adapters.mmaction.utils.config_utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import defaultdict

import pytest
from mmcv.utils import Config

from otx.algorithms.action.adapters.mmaction.utils import (
    patch_config,
    prepare_for_training,
    set_data_classes,
)
from otx.algorithms.common.adapters.mmcv.utils import get_data_cfg
from otx.api.entities.annotation import NullAnnotationSceneEntity
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.model_template import TaskType
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from tests.test_suite.e2e_test_system import e2e_pytest_unit

CLS_CONFIG_NAME = "otx/algorithms/action/configs/classification/x3d/model.py"
DET_CONFIG_NAME = "otx/algorithms/action/configs/detection/x3d_fast_rcnn/model.py"
CLS_CONFIG = Config.fromfile(CLS_CONFIG_NAME)
DET_CONFIG = Config.fromfile(DET_CONFIG_NAME)


@e2e_pytest_unit
def test_patch_config() -> None:
    """Test patch_config function.

    <Steps>
        1. Check error when gives wrong task type
        2. Check work_dir
        3. Check merged data pipeline
        4. Check dataset type
    """

    cls_datapipeline_path = "otx/algorithms/action/configs/classification/x3d/data_pipeline.py"
    work_dir = "OTX-tempdir9104"

    with pytest.raises(NotImplementedError):
        patch_config(CLS_CONFIG, cls_datapipeline_path, work_dir, TaskType.CLASSIFICATION)

    patch_config(CLS_CONFIG, cls_datapipeline_path, work_dir, TaskType.ACTION_CLASSIFICATION)
    assert CLS_CONFIG.work_dir == work_dir
    assert CLS_CONFIG.get("train_pipeline", None)
    for subset in ("train", "val", "test", "unlabeled"):
        cfg = CLS_CONFIG.data.get(subset, None)
        if not cfg:
            continue
        assert cfg.type == "OTXActionClsDataset"

    det_datapipeline_path = "otx/algorithms/action/configs/detection/x3d_fast_rcnn/data_pipeline.py"
    patch_config(DET_CONFIG, det_datapipeline_path, work_dir, TaskType.ACTION_DETECTION)
    assert DET_CONFIG.work_dir == work_dir
    assert DET_CONFIG.get("train_pipeline", None)
    for subset in ("train", "val", "test", "unlabeled"):
        cfg = DET_CONFIG.data.get(subset, None)
        if not cfg:
            continue
        assert cfg.type == "OTXActionDetDataset"


@e2e_pytest_unit
def test_set_data_classes() -> None:
    """Test set_data_classes funciton.

    <Step>
        1. Create sample labels
        2. Check classification label length is appropriate
    """

    labels = [
        LabelEntity(name="0", domain=Domain.ACTION_CLASSIFICATION),
        LabelEntity(name="1", domain=Domain.ACTION_CLASSIFICATION),
        LabelEntity(name="2", domain=Domain.ACTION_CLASSIFICATION),
    ]
    set_data_classes(CLS_CONFIG, labels, TaskType.ACTION_CLASSIFICATION)
    assert CLS_CONFIG.model["cls_head"].num_classes == len(labels)

    labels = [
        LabelEntity(name="0", domain=Domain.ACTION_DETECTION),
        LabelEntity(name="1", domain=Domain.ACTION_DETECTION),
        LabelEntity(name="2", domain=Domain.ACTION_DETECTION),
    ]
    set_data_classes(DET_CONFIG, labels, TaskType.ACTION_DETECTION)
    assert DET_CONFIG.model["roi_head"]["bbox_head"].num_classes == len(labels) + 1
    assert DET_CONFIG.model["roi_head"]["bbox_head"]["topk"] == len(labels) - 1


@e2e_pytest_unit
def test_prepare_for_training() -> None:
    """Test prepare_for_training function.

    <Step>
        1. Create sample DatasetEntity, TimeMonitorCallback, Learngin_cureves
        2. Check config.data.train
        3. Check config.data.val
        4. Check custom_hooks
        5. Check log_config.hooks
    """

    item = DatasetItemEntity(media=Image(file_path="iamge.jpg"), annotation_scene=NullAnnotationSceneEntity())
    dataset = DatasetEntity(items=[item])
    time_monitor = TimeMonitorCallback()
    learning_curves = defaultdict()

    CLS_CONFIG.runner = {}
    CLS_CONFIG.custom_hooks = []
    prepare_for_training(CLS_CONFIG, dataset, dataset, time_monitor, learning_curves)

    assert get_data_cfg(CLS_CONFIG, "train").otx_dataset == dataset
    assert get_data_cfg(CLS_CONFIG, "val").otx_dataset == dataset
    assert (
        CLS_CONFIG.custom_hooks[-1]["type"] == "OTXProgressHook"
        and CLS_CONFIG.custom_hooks[-1]["time_monitor"] == time_monitor
    )
    assert (
        CLS_CONFIG.log_config.hooks[-1]["type"] == "OTXLoggerHook"
        and CLS_CONFIG.log_config.hooks[-1]["curves"] == learning_curves
    )



"""Collection of utils to run common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .callback import (
    InferenceProgressCallback,
    OptimizationProgressCallback,
    TrainingProgressCallback,
)
from .data import get_cls_img_indices, get_image, get_old_new_img_indices
from .utils import UncopiableDefaultDict, get_arg_spec, get_task_class, load_template

__all__ = [
    "get_cls_img_indices",
    "get_old_new_img_indices",
    "TrainingProgressCallback",
    "InferenceProgressCallback",
    "OptimizationProgressCallback",
    "UncopiableDefaultDict",
    "load_template",
    "get_task_class",
    "get_arg_spec",
    "get_image",
]



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import numpy as np
import pytest

from otx.algorithms.common.tasks import BaseTask
from otx.algorithms.segmentation.tasks import SegmentationInferenceTask
from otx.api.configuration.helper import create
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.metrics import Performance, ScoreMetric
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.utils.shape_factory import ShapeFactory
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_SEG_TEMPLATE_DIR,
    generate_otx_dataset,
    init_environment,
)


class TestOTXSegTaskInference:
    @pytest.fixture(autouse=True)
    def setup(self, otx_model, tmp_dir_path) -> None:
        model_template = parse_model_template(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        task_env = init_environment(hyper_parameters, model_template)
        self.output_path = str(tmp_dir_path)
        self.seg_train_task = SegmentationInferenceTask(task_env, output_path=self.output_path)
        self.model = otx_model

    @e2e_pytest_unit
    def test_infer(self, mocker):
        dataset = generate_otx_dataset(5)
        fake_output = {"outputs": {"eval_predictions": np.zeros((5, 1)), "feature_vectors": np.zeros((5, 1))}}
        fake_annotation = [
            Annotation(
                Polygon(points=[Point(0, 0)]),
                id=0,
                labels=[ScoredLabel(LabelEntity(name="fake", domain="SEGMENTATION"), probability=1.0)],
            )
        ]

        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)
        mocker.patch("numpy.transpose")
        mocker.patch("otx.algorithms.segmentation.tasks.inference.create_hard_prediction_from_soft_prediction")
        mocker.patch(
            "otx.algorithms.segmentation.tasks.inference.create_annotation_from_segmentation_map",
            return_value=fake_annotation,
        )
        mocker.patch("otx.algorithms.segmentation.tasks.inference.get_activation_map", return_value=np.zeros((1, 1)))
        mocker.patch.object(ShapeFactory, "shape_produces_valid_crop", return_value=True)

        updated_dataset = self.seg_train_task.infer(dataset, None)

        mock_run_task.assert_called_once()
        for updated in updated_dataset:
            assert updated.annotation_scene.contains_any([LabelEntity(name="fake", domain="SEGMENTATION")])

    @e2e_pytest_unit
    def test_evaluate(self, mocker):
        result_set = ResultSetEntity(
            model=self.model,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.dice.DiceAverage", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics="mDice"
        )
        mocker.patch.object(MetricsHelper, "compute_dice_averaged_over_pixels", return_value=fake_metrics)
        self.seg_train_task.evaluate(result_set)

        assert result_set.performance.score.value == 0.1

    @e2e_pytest_unit
    def test_export(self, mocker):
        fake_output = {"outputs": {"bin": None, "xml": None}}
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)

        with pytest.raises(RuntimeError):
            self.seg_train_task.export(ExportType.OPENVINO, self.model)
            mock_run_task.assert_called_once()

    @e2e_pytest_unit
    def test_export_with_model_files(self, mocker):
        with open(f"{self.output_path}/model.xml", "wb") as f:
            f.write(b"foo")
        with open(f"{self.output_path}/model.bin", "wb") as f:
            f.write(b"bar")

        fake_output = {"outputs": {"bin": f"{self.output_path}/model.xml", "xml": f"{self.output_path}/model.bin"}}
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)
        self.seg_train_task.export(ExportType.OPENVINO, self.model)

        mock_run_task.assert_called_once()
        assert self.model.get_data("openvino.bin")
        assert self.model.get_data("openvino.xml")

    @e2e_pytest_unit
    def test_unload(self, mocker):
        mock_cleanup = mocker.patch.object(BaseTask, "cleanup")
        self.seg_train_task.unload()

        mock_cleanup.assert_called_once()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools

import torch
from mmdet.models.builder import DETECTORS
from mmdet.models.detectors.yolox import YOLOX

from otx.mpa.deploy.utils import is_mmdeploy_enabled
from otx.mpa.modules.utils.task_adapt import map_class_names
from otx.mpa.utils.logger import get_logger

from .l2sp_detector_mixin import L2SPDetectorMixin
from .sam_detector_mixin import SAMDetectorMixin

logger = get_logger()


@DETECTORS.register_module()
class CustomYOLOX(SAMDetectorMixin, L2SPDetectorMixin, YOLOX):
    """SAM optimizer & L2SP regularizer enabled custom YOLOX"""

    def __init__(self, *args, task_adapt=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Hook for class-sensitive weight loading
        if task_adapt:
            self._register_load_state_dict_pre_hook(
                functools.partial(
                    self.load_state_dict_pre_hook,
                    self,  # model
                    task_adapt["dst_classes"],  # model_classes
                    task_adapt["src_classes"],  # chkpt_classes
                )
            )

    def forward_train(self, img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=None, **kwargs):
        return super().forward_train(img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=gt_bboxes_ignore)

    @staticmethod
    def load_state_dict_pre_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading"""
        logger.info(f"----------------- CustomYOLOX.load_state_dict_pre_hook() called w/ prefix: {prefix}")

        # Dst to src mapping index
        model_classes = list(model_classes)
        chkpt_classes = list(chkpt_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes} ({model2chkpt})")

        model_dict = model.state_dict()
        param_names = [
            "bbox_head.multi_level_conv_cls.0.weight",
            "bbox_head.multi_level_conv_cls.0.bias",
            "bbox_head.multi_level_conv_cls.1.weight",
            "bbox_head.multi_level_conv_cls.1.bias",
            "bbox_head.multi_level_conv_cls.2.weight",
            "bbox_head.multi_level_conv_cls.2.bias",
        ]
        for model_name in param_names:
            chkpt_name = prefix + model_name
            if model_name not in model_dict or chkpt_name not in chkpt_dict:
                logger.info(f"Skipping weight copy: {chkpt_name}")
                continue

            # Mix weights
            model_param = model_dict[model_name].clone()
            chkpt_param = chkpt_dict[chkpt_name]
            for m, c in enumerate(model2chkpt):
                if c >= 0:
                    model_param[m].copy_(chkpt_param[c])

            # Replace checkpoint weight by mixed weights
            chkpt_dict[chkpt_name] = model_param

    def onnx_export(self, img, img_metas, with_nms=True):
        """Test function without test time augmentation.

        Args:
            img (torch.Tensor): input images.
            img_metas (list[dict]): List of image information.

        Returns:
            tuple[Tensor, Tensor]: dets of shape [N, num_det, 5]
                and class labels of shape [N, num_det].
        """
        x = self.extract_feat(img)
        outs = self.bbox_head(x)
        # get origin input shape to support onnx dynamic shape

        # get shape as tensor
        img_shape = torch._shape_as_tensor(img)[2:]
        img_metas[0]["img_shape_for_onnx"] = img_shape
        # get pad input shape to support onnx dynamic shape for exporting
        # `CornerNet` and `CentripetalNet`, which 'pad_shape' is used
        # for inference
        img_metas[0]["pad_shape_for_onnx"] = img_shape

        if len(outs) == 2:
            # add dummy score_factor
            outs = (*outs, None)

        # FIXME: mmdet does not support yolox onnx export for now
        # This is a temporary workaround
        # https://github.com/open-mmlab/mmdetection/issues/6487
        det_bboxes, det_labels = self.bbox_head.get_bboxes(*outs, img_metas)[0]

        return det_bboxes, det_labels


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER, mark
    from mmdeploy.utils import is_dynamic_shape

    from otx.mpa.modules.hooks.recording_forward_hooks import (
        DetSaliencyMapHook,
        FeatureVectorHook,
    )

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.detectors.custom_yolox_detector.CustomYOLOX.simple_test"
    )
    def custom_yolox__simple_test(ctx, self, img, img_metas, **kwargs):
        feat = self.extract_feat(img)
        outs = self.bbox_head(feat)
        bbox_results = self.bbox_head.get_bboxes(*outs, img_metas=img_metas, cfg=self.test_cfg, **kwargs)

        if ctx.cfg["dump_features"]:
            feature_vector = FeatureVectorHook.func(feat)
            cls_scores = outs[0]
            saliency_map = DetSaliencyMapHook(self).func(cls_scores, cls_scores_provided=True)
            return (*bbox_results, feature_vector, saliency_map)

        return bbox_results

    @mark("custom_yolox_forward", inputs=["input"], outputs=["dets", "labels", "feats", "saliencies"])
    def __forward_impl(ctx, self, img, img_metas, **kwargs):
        assert isinstance(img, torch.Tensor)

        deploy_cfg = ctx.cfg
        is_dynamic_flag = is_dynamic_shape(deploy_cfg)
        # get origin input shape as tensor to support onnx dynamic shape
        img_shape = torch._shape_as_tensor(img)[2:]
        if not is_dynamic_flag:
            img_shape = [int(val) for val in img_shape]
        img_metas[0]["img_shape"] = img_shape
        return self.simple_test(img, img_metas, **kwargs)

    @FUNCTION_REWRITER.register_rewriter("otx.mpa.modules.models.detectors.custom_yolox_detector.CustomYOLOX.forward")
    def custom_yolox__forward(ctx, self, img, img_metas=None, return_loss=False, **kwargs):
        if img_metas is None:
            img_metas = [{}]
        else:
            assert len(img_metas) == 1, "do not support aug_test"
            img_metas = img_metas[0]

        if isinstance(img, list):
            img = img[0]

        return __forward_impl(ctx, self, img, img_metas=img_metas, **kwargs)



"""Data Pipeline for Semi-SL model of Segmentation Task."""

__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__img_scale = (544, 544)
__crop_size = (512, 512)

__common_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations"),
    dict(type="Resize", img_scale=__img_scale, ratio_range=(0.5, 2.0), keep_ratio=False),
    dict(type="RandomCrop", crop_size=__crop_size, cat_max_ratio=0.75),
    dict(type="RandomFlip", prob=0.5, direction="horizontal"),
    dict(type="RandomRotate", prob=0.5, degree=30, pad_val=0, seg_pad_val=255),
]

train_pipeline = [
    *__common_pipeline,
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="Pad", size=__crop_size, pad_val=0, seg_pad_val=255),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_semantic_seg"]),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_scale,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]

unlabeled_pipeline = [
    *__common_pipeline,
    dict(type="Pad", size=__crop_size, pad_val=0, seg_pad_val=255),
    dict(type="BranchImage", key_map=dict(img="ul_w_img")),
    dict(
        type="ProbCompose",
        probs=[0.7, 0.3],
        transforms=[
            dict(
                type="PhotoMetricDistortion",
                brightness_delta=32,
                contrast_range=(0.5, 1.5),
                saturation_range=(0.5, 1.5),
                hue_delta=18,
            ),
            dict(type="RGB2Gray"),
        ],
    ),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "ul_w_img"]),
]

# TODO (Soobee) : Remove Repeatdataset in data config
data = dict(
    train=dict(dataset=dict(pipeline=train_pipeline)),
    val=dict(pipeline=test_pipeline),
    test=dict(pipeline=test_pipeline),
    unlabeled=dict(pipeline=unlabeled_pipeline),
)



"""Unit Test for otx.algorithms.detection.adapters.mmdet.utils.builder."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import pytest

from otx.algorithms.detection.adapters.mmdet.utils import build_detector
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    DEFAULT_DET_MODEL_CONFIG_PATH,
    DEFAULT_ISEG_MODEL_CONFIG_PATH,
)


@e2e_pytest_unit
@pytest.mark.parametrize("model_cfg", [DEFAULT_DET_MODEL_CONFIG_PATH, DEFAULT_ISEG_MODEL_CONFIG_PATH])
@pytest.mark.parametrize("cfg_options", [None, MPAConfig()])
def test_build_detector(model_cfg, cfg_options):
    """Test build_detector function."""
    cfg = MPAConfig.fromfile(model_cfg)
    model = build_detector(cfg, checkpoint=cfg.load_from, cfg_options=cfg_options)
    assert model is not None



"""Model wrapper file for openvino."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

from typing import Any, Dict

import numpy as np

from otx.api.utils.argument_checks import check_input_parameters_type

try:
    from openvino.model_zoo.model_api.models.classification import Classification
    from openvino.model_zoo.model_api.models.types import BooleanValue, DictValue
except ImportError:
    import warnings

    warnings.warn("ModelAPI was not found.")


class OTXClassification(Classification):
    """OTX classification class for openvino."""

    __model__ = "otx_classification"

    def __init__(self, model_adapter, configuration=None, preload=False):
        super().__init__(model_adapter, configuration, preload)
        if self.hierarchical:
            logits_range_dict = self.multihead_class_info.get("head_idx_to_logits_range", False)
            if logits_range_dict:  # json allows only string key, revert to integer.
                self.multihead_class_info["head_idx_to_logits_range"] = {
                    int(k): v for k, v in logits_range_dict.items()
                }

    @classmethod
    def parameters(cls):
        """Parameters."""
        parameters = super().parameters()
        parameters["resize_type"].update_default_value("standard")
        parameters.update(
            {
                "multilabel": BooleanValue(default_value=False),
                "hierarchical": BooleanValue(default_value=False),
                "multihead_class_info": DictValue(default_value={}),
            }
        )

        return parameters

    def _check_io_number(self, number_of_inputs, number_of_outputs):
        pass

    def _get_outputs(self):
        layer_name = "logits"
        for name, meta in self.outputs.items():
            if "logits" in meta.names:
                layer_name = name
        layer_shape = self.outputs[layer_name].shape

        if len(layer_shape) != 2 and len(layer_shape) != 4:
            raise RuntimeError("The Classification model wrapper supports topologies only with 2D or 4D output")
        if len(layer_shape) == 4 and (layer_shape[2] != 1 or layer_shape[3] != 1):
            raise RuntimeError(
                "The Classification model wrapper supports topologies only with 4D "
                "output which has last two dimensions of size 1"
            )
        if self.labels:
            if layer_shape[1] == len(self.labels) + 1:
                self.labels.insert(0, "other")
                self.logger.warning("\tInserted 'other' label as first.")
            if layer_shape[1] != len(self.labels):
                raise RuntimeError(
                    "Model's number of classes and parsed "
                    f"labels must match ({layer_shape[1]} != {len(self.labels)})"
                )
        return layer_name

    @check_input_parameters_type()
    def postprocess(self, outputs: Dict[str, np.ndarray], meta: Dict[str, Any]):  # pylint: disable=unused-argument
        """Post-process."""
        logits = outputs[self.out_layer_name].squeeze()
        if self.multilabel:
            return get_multilabel_predictions(logits)
        if self.hierarchical:
            return get_hierarchical_predictions(logits, self.multihead_class_info)

        return get_multiclass_predictions(logits)

    # pylint: disable=unused-argument
    @check_input_parameters_type()
    def postprocess_aux_outputs(self, outputs: Dict[str, np.ndarray], metadata: Dict[str, Any]):
        """Post-process for auxiliary outputs."""
        logits = outputs[self.out_layer_name].squeeze()
        if self.multilabel:
            probs = sigmoid_numpy(logits)
        elif self.hierarchical:
            probs = activate_multihead_output(logits, self.multihead_class_info)
        else:
            probs = softmax_numpy(logits)
        act_score = float(np.max(probs) - np.min(probs))

        if "saliency_map" in outputs:
            saliency_map = outputs["saliency_map"][0]
            repr_vector = outputs["feature_vector"].reshape(-1)
        else:
            saliency_map, repr_vector = None, None

        return probs, saliency_map, repr_vector, act_score


@check_input_parameters_type()
def sigmoid_numpy(x: np.ndarray):
    """Sigmoid numpy."""
    return 1.0 / (1.0 + np.exp(-1.0 * x))


@check_input_parameters_type()
def softmax_numpy(x: np.ndarray, eps: float = 1e-9):
    """Softmax numpy."""
    x = np.exp(x - np.max(x))
    return x / (np.sum(x) + eps)


@check_input_parameters_type()
def activate_multihead_output(logits: np.ndarray, multihead_class_info: dict):
    """Activate multi-head output."""
    for i in range(multihead_class_info["num_multiclass_heads"]):
        logits_begin, logits_end = multihead_class_info["head_idx_to_logits_range"][i]
        logits[logits_begin:logits_end] = softmax_numpy(logits[logits_begin:logits_end])

    if multihead_class_info["num_multilabel_classes"]:
        logits_begin = multihead_class_info["num_single_label_classes"]
        logits[logits_begin:] = sigmoid_numpy(logits[logits_begin:])

    return logits


@check_input_parameters_type()
def get_hierarchical_predictions(
    logits: np.ndarray, multihead_class_info: dict, pos_thr: float = 0.5, activate: bool = True
):
    """Get hierarchical predictions."""
    predicted_labels = []
    for i in range(multihead_class_info["num_multiclass_heads"]):
        logits_begin, logits_end = multihead_class_info["head_idx_to_logits_range"][i]
        head_logits = logits[logits_begin:logits_end]
        if activate:
            head_logits = softmax_numpy(head_logits)
        j = np.argmax(head_logits)
        label_str = multihead_class_info["all_groups"][i][j]
        predicted_labels.append((multihead_class_info["label_to_idx"][label_str], head_logits[j]))

    if multihead_class_info["num_multilabel_classes"]:
        logits_begin = multihead_class_info["num_single_label_classes"]
        head_logits = logits[logits_begin:]
        if activate:
            head_logits = sigmoid_numpy(head_logits)

        for i in range(head_logits.shape[0]):
            if head_logits[i] > pos_thr:
                label_str = multihead_class_info["all_groups"][multihead_class_info["num_multiclass_heads"] + i][0]
                predicted_labels.append((multihead_class_info["label_to_idx"][label_str], head_logits[i]))

    return predicted_labels


@check_input_parameters_type()
def get_multiclass_predictions(logits: np.ndarray, activate: bool = True):
    """Get multiclass predictions."""
    index = np.argmax(logits)
    if activate:
        logits = softmax_numpy(logits)
    return [(index, logits[index])]


@check_input_parameters_type()
def get_multilabel_predictions(logits: np.ndarray, pos_thr: float = 0.5, activate: bool = True):
    """Get multilabel predictions."""
    if activate:
        logits = sigmoid_numpy(logits)
    scores = []
    indices = []
    for i in range(logits.shape[0]):
        if logits[i] > pos_thr:
            indices.append(i)
            scores.append(logits[i])

    return list(zip(indices, scores))



_base_ = "./model.py"

model = dict(
    type="SupConClassifier",
    task="classification",
    pretrained=None,
    backbone=dict(),
    neck=dict(type="GlobalAveragePooling"),
    head=dict(
        type="SupConClsHead",
        num_classes=10,
        in_channels=-1,
        aux_mlp=dict(hid_channels=0, out_channels=1024),
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
        aux_loss=dict(
            type="BarlowTwinsLoss",
            off_diag_penality=1.0 / 128.0,
            loss_weight=1.0,
        ),
    ),
)

checkpoint_config = dict(type="CheckpointHookWithValResults")



"""This module contains the base class for deep learning image-based tasks."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc

from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.training_interface import ITrainingTask


class ImageDeepLearningTask(IInferenceTask, ITrainingTask, IEvaluationTask, metaclass=abc.ABCMeta):
    """A base class for a deep learning image-based tasks.

    This class inherits from ITask, ITraining, IComputesPerformance and IReporting.

    Example:
        A YOLO detection task.

        >>> class YOLODetection(ImageDeepLearningTask):
        ...     pass
    """
